import os
import functools
import inspect
import os
import sys
import traceback
from datetime import datetime

import metaflow.tracing as tracing
from metaflow._vendor import click

from . import decorators, lint, metaflow_version, parameters, plugins
from .cli_args import cli_args
from .cli_components.utils import LazyGroup, LazyPluginCommandCollection
from .datastore import FlowDataStore, TaskDataStoreSet
from .debug import debug
from .exception import CommandException, MetaflowException
from .flowspec import _FlowState
from .graph import FlowGraph
from .metaflow_config import (
    DEFAULT_DATASTORE,
    DEFAULT_DECOSPECS,
    DEFAULT_ENVIRONMENT,
    DEFAULT_EVENT_LOGGER,
    DEFAULT_METADATA,
    DEFAULT_MONITOR,
    DEFAULT_PACKAGE_SUFFIXES,
)
from .metaflow_current import current
from .metaflow_profile import from_start
from metaflow.system import _system_monitor, _system_logger
from .metaflow_environment import MetaflowEnvironment
from .packaging_sys import MetaflowCodeContent
from .plugins import (
    DATASTORES,
    ENVIRONMENTS,
    LOGGING_SIDECARS,
    METADATA_PROVIDERS,
    MONITOR_SIDECARS,
)
from .pylint_wrapper import PyLint
from .R import metaflow_r_version, use_r
from .util import get_latest_run_id, resolve_identity, decompress_list
from .user_configs.config_options import LocalFileInput, config_options
from .user_configs.config_parameters import ConfigValue

ERASE_TO_EOL = "\033[K"
HIGHLIGHT = "red"
INDENT = " " * 4

LOGGER_TIMESTAMP = "magenta"
LOGGER_COLOR = "green"
LOGGER_BAD_COLOR = "red"


def echo_dev_null(*args, **kwargs):
    pass


def echo_always(line, **kwargs):
    if kwargs.pop("wrap", False):
        import textwrap

        indent_str = INDENT if kwargs.get("indent", None) else ""
        effective_width = 80 - len(indent_str)
        wrapped = textwrap.wrap(line, width=effective_width, break_long_words=False)
        line = "\n".join(indent_str + l for l in wrapped)
        kwargs["indent"] = False

    kwargs["err"] = kwargs.get("err", True)
    if kwargs.pop("indent", None):
        line = "\n".join(INDENT + x for x in line.splitlines())
    if "nl" not in kwargs or kwargs["nl"]:
        line += ERASE_TO_EOL
    top = kwargs.pop("padding_top", None)
    bottom = kwargs.pop("padding_bottom", None)
    highlight = kwargs.pop("highlight", HIGHLIGHT)
    if top:
        click.secho(ERASE_TO_EOL, **kwargs)

    hl_bold = kwargs.pop("highlight_bold", True)
    nl = kwargs.pop("nl", True)
    fg = kwargs.pop("fg", None)
    bold = kwargs.pop("bold", False)
    kwargs["nl"] = False
    hl = True
    nobold = kwargs.pop("no_bold", False)
    if nobold:
        click.secho(line, **kwargs)
    else:
        for span in line.split("*"):
            if hl:
                hl = False
                kwargs["fg"] = fg
                kwargs["bold"] = bold
                click.secho(span, **kwargs)
            else:
                hl = True
                kwargs["fg"] = highlight
                kwargs["bold"] = hl_bold
                click.secho(span, **kwargs)
    if nl:
        kwargs["nl"] = True
        click.secho("", **kwargs)
    if bottom:
        click.secho(ERASE_TO_EOL, **kwargs)


def logger(body="", system_msg=False, head="", bad=False, timestamp=True, nl=True):
    if timestamp:
        if timestamp is True:
            dt = datetime.now()
        else:
            dt = timestamp
        tstamp = dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        click.secho(tstamp + " ", fg=LOGGER_TIMESTAMP, nl=False)
    if head:
        click.secho(head, fg=LOGGER_COLOR, nl=False)
    click.secho(body, bold=system_msg, fg=LOGGER_BAD_COLOR if bad else None, nl=nl)


@click.group(
    cls=LazyGroup,
    lazy_subcommands={
        "init": "metaflow.cli_components.init_cmd.init",
        "dump": "metaflow.cli_components.dump_cmd.dump",
        "step": "metaflow.cli_components.step_cmd.step",
        "run": "metaflow.cli_components.run_cmds.run",
        "resume": "metaflow.cli_components.run_cmds.resume",
        "spin": "metaflow.cli_components.run_cmds.spin",
        "spin-step": "metaflow.cli_components.step_cmd.spin_step",
    },
)
def cli(ctx):
    pass


@cli.command(help="Check that the flow is valid (default).")
@click.option(
    "--warnings/--no-warnings",
    default=False,
    show_default=True,
    help="Show all Pylint warnings, not just errors.",
)
@click.pass_obj
def check(obj, warnings=False):
    if obj.is_quiet:
        echo = echo_dev_null
    else:
        echo = echo_always
    _check(
        echo, obj.graph, obj.flow, obj.environment, pylint=obj.pylint, warnings=warnings
    )
    fname = inspect.getfile(obj.flow.__class__)
    echo(
        "\n*'{cmd} show'* shows a description of this flow.\n"
        "*'{cmd} run'* runs the flow locally.\n"
        "*'{cmd} help'* shows all available commands and options.\n".format(cmd=fname),
        highlight="magenta",
        highlight_bold=False,
    )


@cli.command(help="Show structure of the flow.")
@click.pass_obj
def show(obj):
    echo_always("\n%s" % obj.graph.doc)
    for node_name in obj.graph.sorted_nodes:
        echo_always("")
        node = obj.graph[node_name]
        for deco in node.decorators:
            echo_always("@%s" % deco.name, err=False)
        for deco in node.wrappers:
            echo_always("@%s" % deco.decorator_name, err=False)
        echo_always("Step *%s*" % node.name, err=False)
        echo_always(node.doc if node.doc else "?", indent=True, err=False)
        if node.type != "end":
            echo_always(
                "*=>* %s" % ", ".join("*%s*" % n for n in node.out_funcs),
                indent=True,
                highlight="magenta",
                highlight_bold=False,
                err=False,
            )
    echo_always("")


@cli.command(help="Show all available commands.")
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())


@cli.command(help="Output internal state of the flow graph.")
@click.option("--json", is_flag=True, help="Output the flow graph in JSON format.")
@click.pass_obj
def output_raw(obj, json):
    if json:
        import json as _json

        _msg = "Internal representation of the flow in JSON format:"
        _graph_dict, _graph_struct = obj.graph.output_steps()
        _graph = _json.dumps(
            dict(graph=_graph_dict, graph_structure=_graph_struct), indent=4
        )
    else:
        _graph = str(obj.graph)
        _msg = "Internal representation of the flow:"
    echo_always(_msg, fg="magenta", bold=False)
    echo_always(_graph, err=False)


@cli.command(help="Visualize the flow with Graphviz.")
@click.pass_obj
def output_dot(obj):
    echo_always("Visualizing the flow as a GraphViz graph", fg="magenta", bold=False)
    echo_always(
        "Try piping the output to 'dot -Tpng -o graph.png' to produce "
        "an actual image.",
        indent=True,
    )
    echo_always(obj.graph.output_dot(), err=False)


@cli.command(help="Print the Metaflow version")
@click.pass_obj
def version(obj):
    echo_always(obj.version)


# NOTE: add_decorator_options should be TL because it checks to make sure
# that no option conflict with the ones below
@decorators.add_decorator_options
@config_options
@click.command(
    cls=LazyPluginCommandCollection,
    sources=[cli],
    lazy_sources=plugins.get_plugin_cli_path(),
    invoke_without_command=True,
)
# Quiet is eager to make sure it is available when processing --config options since
# we need it to construct a context to pass to any DeployTimeField for the default
# value.
@click.option(
    "--quiet/--not-quiet",
    show_default=True,
    default=False,
    help="Suppress unnecessary messages",
    is_eager=True,
)
@click.option(
    "--metadata",
    default=DEFAULT_METADATA,
    show_default=True,
    type=click.Choice([m.TYPE for m in METADATA_PROVIDERS]),
    help="Metadata service type",
)
@click.option(
    "--environment",
    default=DEFAULT_ENVIRONMENT,
    show_default=True,
    type=click.Choice(["local"] + [m.TYPE for m in ENVIRONMENTS]),
    help="Execution environment type",
)
@click.option(
    "--force-rebuild-environments/--no-force-rebuild-environments",
    is_flag=True,
    default=False,
    hidden=True,
    type=bool,
    help="Explicitly rebuild the execution environments",
)
# See comment for --quiet
@click.option(
    "--datastore",
    default=DEFAULT_DATASTORE,
    show_default=True,
    type=click.Choice([d.TYPE for d in DATASTORES]),
    help="Data backend type",
    is_eager=True,
)
@click.option("--datastore-root", help="Root path for datastore")
@click.option(
    "--package-suffixes",
    help="A comma-separated list of file suffixes to include in the code package.",
    default=DEFAULT_PACKAGE_SUFFIXES,
    show_default=True,
)
@click.option(
    "--with",
    "decospecs",
    multiple=True,
    help="Add a decorator to all steps. You can specify this option "
    "multiple times to attach multiple decorators in steps.",
)
@click.option(
    "--pylint/--no-pylint",
    default=True,
    show_default=True,
    help="Run Pylint on the flow if pylint is installed.",
)
@click.option(
    "--event-logger",
    default=DEFAULT_EVENT_LOGGER,
    show_default=True,
    type=click.Choice(LOGGING_SIDECARS),
    help="type of event logger used",
)
@click.option(
    "--monitor",
    default=DEFAULT_MONITOR,
    show_default=True,
    type=click.Choice(MONITOR_SIDECARS),
    help="Monitoring backend type",
)
@click.option(
    "--local-config-file",
    type=LocalFileInput(exists=True, readable=True, dir_okay=False, resolve_path=True),
    required=False,
    default=None,
    help="A filename containing the dumped configuration values. Internal use only.",
    hidden=True,
    is_eager=True,
)
@click.option(
    "--mode",
    type=click.Choice(["spin"]),
    default=None,
    help="Execution mode for metaflow CLI commands. Use 'spin' to enable "
    "spin metadata and spin datastore for executions",
)
@click.pass_context
def start(
    ctx,
    quiet=False,
    metadata=None,
    environment=None,
    force_rebuild_environments=False,
    datastore=None,
    datastore_root=None,
    decospecs=None,
    package_suffixes=None,
    pylint=None,
    event_logger=None,
    monitor=None,
    local_config_file=None,
    config=None,
    config_value=None,
    mode=None,
    **deco_options
):
    if quiet:
        echo = echo_dev_null
    else:
        echo = echo_always

    ctx.obj.version = metaflow_version.get_version()
    version = ctx.obj.version
    if use_r():
        version = metaflow_r_version()

    from_start("MetaflowCLI: Starting")
    echo("Metaflow %s" % version, fg="magenta", bold=True, nl=False)
    echo(" executing *%s*" % ctx.obj.flow.name, fg="magenta", nl=False)
    echo(" for *%s*" % resolve_identity(), fg="magenta")

    # Check if we need to setup the distribution finder (if running )
    dist_info = MetaflowCodeContent.get_distribution_finder()
    if dist_info:
        sys.meta_path.append(dist_info)

    # Setup the context
    cli_args._set_top_kwargs(ctx.params)
    ctx.obj.echo = echo
    ctx.obj.echo_always = echo_always
    ctx.obj.is_quiet = quiet
    ctx.obj.logger = logger
    ctx.obj.pylint = pylint
    ctx.obj.check = functools.partial(_check, echo)
    ctx.obj.top_cli = cli
    ctx.obj.package_suffixes = package_suffixes.split(",")
    ctx.obj.spin_mode = mode == "spin"

    ctx.obj.datastore_impl = [d for d in DATASTORES if d.TYPE == datastore][0]

    if datastore_root is None:
        datastore_root = ctx.obj.datastore_impl.get_datastore_root_from_config(
            ctx.obj.echo
        )
    if datastore_root is None:
        raise CommandException(
            "Could not find the location of the datastore -- did you correctly set the "
            "METAFLOW_DATASTORE_SYSROOT_%s environment variable?" % datastore.upper()
        )

    ctx.obj.datastore_impl.datastore_root = datastore_root

    FlowDataStore.default_storage_impl = ctx.obj.datastore_impl

    # At this point, we are able to resolve the user-configuration options so we can
    # process all those decorators that the user added that will modify the flow based
    # on those configurations. It is important to do this as early as possible since it
    # actually modifies the flow itself

    # When we process the options, the first one processed will return None and the
    # second one processed will return the actual options. The order of processing
    # depends on what (and in what order) the user specifies on the command line.
    config_options = config or config_value

    if (
        hasattr(ctx, "saved_args")
        and ctx.saved_args
        and ctx.saved_args[0] == "resume"
        and getattr(ctx.obj, "has_config_options", False)
    ):
        # In the case of resume, we actually need to load the configurations
        # from the resumed run to process them. This can be slightly onerous so check
        # if we need to in the first place
        if getattr(ctx.obj, "has_cl_config_options", False):
            raise click.UsageError(
                "Cannot specify --config or --config-value with 'resume'"
            )
        # We now load the config artifacts from the original run id
        run_id = None
        try:
            idx = ctx.saved_args.index("--origin-run-id")
        except ValueError:
            idx = -1
        if idx >= 0:
            run_id = ctx.saved_args[idx + 1]
        else:
            run_id = get_latest_run_id(ctx.obj.echo, ctx.obj.flow.name)
        if run_id is None:
            raise CommandException(
                "A previous run id was not found. Specify --origin-run-id."
            )
        # We get the name of the parameters we need to load from the datastore -- these
        # are accessed using the *variable* name and not necessarily the *parameter* name
        config_var_names = []
        config_param_names = []
        for name, param in ctx.obj.flow._get_parameters():
            if not param.IS_CONFIG_PARAMETER:
                continue
            config_var_names.append(name)
            config_param_names.append(param.name)

        # We just need a task datastore that will be thrown away -- we do this so
        # we don't have to create the logger, monitor, etc.
        debug.userconf_exec("Loading config parameters from run %s" % run_id)
        for d in TaskDataStoreSet(
            FlowDataStore(ctx.obj.flow.name),
            run_id,
            steps=["_parameters"],
            prefetch_data_artifacts=config_var_names,
        ):
            param_ds = d

        # We can now set the the CONFIGS value in the flow properly. This will overwrite
        # anything that may have been passed in by default and we will use exactly what
        # the original flow had. Note that these are accessed through the parameter name
        ctx.obj.flow._flow_state[_FlowState.CONFIGS].clear()
        d = ctx.obj.flow._flow_state[_FlowState.CONFIGS]
        for param_name, var_name in zip(config_param_names, config_var_names):
            val = param_ds[var_name]
            debug.userconf_exec("Loaded config %s as: %s" % (param_name, val))
            d[param_name] = val

    elif getattr(ctx.obj, "delayed_config_exception", None):
        # If we are not doing a resume, any exception we had parsing configs needs to
        # be raised. For resume, since we ignore those options, we ignore the error.
        raise ctx.obj.delayed_config_exception

    # Init all values in the flow mutators and then process them
    for decorator in ctx.obj.flow._flow_state.get(_FlowState.FLOW_MUTATORS, []):
        decorator.external_init()

    new_cls = ctx.obj.flow._process_config_decorators(config_options)
    if new_cls:
        ctx.obj.flow = new_cls(use_cli=False)

    ctx.obj.graph = ctx.obj.flow._graph

    ctx.obj.environment = [
        e for e in ENVIRONMENTS + [MetaflowEnvironment] if e.TYPE == environment
    ][0](ctx.obj.flow)
    # set force rebuild flag for environments that support it.
    ctx.obj.environment._force_rebuild = force_rebuild_environments
    ctx.obj.environment.validate_environment(ctx.obj.logger, datastore)
    ctx.obj.event_logger = LOGGING_SIDECARS[event_logger](
        flow=ctx.obj.flow, env=ctx.obj.environment
    )
    ctx.obj.monitor = MONITOR_SIDECARS[monitor](
        flow=ctx.obj.flow, env=ctx.obj.environment
    )
    ctx.obj.metadata = [m for m in METADATA_PROVIDERS if m.TYPE == metadata][0](
        ctx.obj.environment, ctx.obj.flow, ctx.obj.event_logger, ctx.obj.monitor
    )

    ctx.obj.flow_datastore = FlowDataStore(
        ctx.obj.flow.name,
        ctx.obj.environment,
        ctx.obj.metadata,
        ctx.obj.event_logger,
        ctx.obj.monitor,
    )

    ctx.obj.config_options = config_options
    ctx.obj.is_spin = False
    ctx.obj.skip_decorators = False

    # Override values for spin steps, or if we are in spin mode
    if (
        hasattr(ctx, "saved_args")
        and ctx.saved_args
        and "spin" in ctx.saved_args[0]
        or ctx.obj.spin_mode
    ):
        # To minimize side effects for spin, we will only use the following:
        # - local metadata provider,
        # - local datastore,
        # - local environment,
        # - null event logger,
        # - null monitor
        ctx.obj.is_spin = True
        if "--skip-decorators" in ctx.saved_args:
            ctx.obj.skip_decorators = True

        ctx.obj.event_logger = LOGGING_SIDECARS["nullSidecarLogger"](
            flow=ctx.obj.flow, env=ctx.obj.environment
        )
        ctx.obj.monitor = MONITOR_SIDECARS["nullSidecarMonitor"](
            flow=ctx.obj.flow, env=ctx.obj.environment
        )
        # Use spin metadata, spin datastore, and spin datastore root
        ctx.obj.metadata = [m for m in METADATA_PROVIDERS if m.TYPE == "spin"][0](
            ctx.obj.environment, ctx.obj.flow, ctx.obj.event_logger, ctx.obj.monitor
        )
        ctx.obj.datastore_impl = [d for d in DATASTORES if d.TYPE == "spin"][0]
        datastore_root = ctx.obj.datastore_impl.get_datastore_root_from_config(
            ctx.obj.echo, create_on_absent=True
        )
        ctx.obj.datastore_impl.datastore_root = datastore_root

        ctx.obj.flow_datastore = FlowDataStore(
            ctx.obj.flow.name,
            ctx.obj.environment,  # Same environment as run/resume
            ctx.obj.metadata,  # local metadata
            ctx.obj.event_logger,  # null event logger
            ctx.obj.monitor,  # null monitor
            storage_impl=ctx.obj.datastore_impl,
        )

    # Start event logger and monitor
    ctx.obj.event_logger.start()
    _system_logger.init_system_logger(ctx.obj.flow.name, ctx.obj.event_logger)

    ctx.obj.monitor.start()
    _system_monitor.init_system_monitor(ctx.obj.flow.name, ctx.obj.monitor)

    decorators._init(ctx.obj.flow)

    # It is important to initialize flow decorators early as some of the
    # things they provide may be used by some of the objects initialized after.
    decorators._init_flow_decorators(
        ctx.obj.flow,
        ctx.obj.graph,
        ctx.obj.environment,
        ctx.obj.flow_datastore,
        ctx.obj.metadata,
        ctx.obj.logger,
        echo,
        deco_options,
        ctx.obj.is_spin,
        ctx.obj.skip_decorators,
    )

    # In the case of run/resume/spin, we will want to apply the TL decospecs
    # *after* the run decospecs so that they don't take precedence. In other
    # words, for the same decorator, we want `myflow.py run --with foo` to
    # take precedence over any other `foo` decospec

    # Note that top-level decospecs are used primarily with non run/resume
    # options as well as with the airflow/argo/sfn integrations which pass
    # all the decospecs (the ones from top-level but also the ones from the
    # run/resume level) through the tl decospecs.
    ctx.obj.tl_decospecs = list(decospecs or [])

    # initialize current and parameter context for deploy-time parameters
    current._set_env(flow=ctx.obj.flow, is_running=False)
    parameters.set_parameter_context(
        ctx.obj.flow.name,
        ctx.obj.echo,
        ctx.obj.flow_datastore,
        {
            k: ConfigValue(v) if v is not None else None
            for k, v in ctx.obj.flow.__class__._flow_state.get(
                _FlowState.CONFIGS, {}
            ).items()
        },
    )

    if (
        hasattr(ctx, "saved_args")
        and ctx.saved_args
        and ctx.saved_args[0] not in ("run", "resume", "spin")
    ):
        # run/resume/spin are special cases because they can add more decorators with --with,
        # so they have to take care of themselves.
        all_decospecs = ctx.obj.tl_decospecs + list(
            ctx.obj.environment.decospecs() or []
        )

        # We add the default decospecs for everything except init and step since in those
        # cases, the decospecs will already have been handled by either a run/resume
        # or a scheduler setting them up in their own way.
        if ctx.saved_args[0] not in ("step", "init"):
            all_decospecs += DEFAULT_DECOSPECS.split()
        elif ctx.saved_args[0] == "spin-step":
            # If we are in spin-args, we will not attach any decorators
            all_decospecs = []
        if all_decospecs:
            decorators._attach_decorators(ctx.obj.flow, all_decospecs)
            decorators._init(ctx.obj.flow)
            # Regenerate graph if we attached more decorators
            ctx.obj.flow.__class__._init_graph()
            ctx.obj.graph = ctx.obj.flow._graph

        decorators._init_step_decorators(
            ctx.obj.flow,
            ctx.obj.graph,
            ctx.obj.environment,
            ctx.obj.flow_datastore,
            ctx.obj.logger,
            # The last two arguments are only used for spin steps
            ctx.obj.is_spin,
            ctx.obj.skip_decorators,
        )

        # Check the graph again (mutators may have changed it)
        ctx.obj.graph = ctx.obj.flow._graph

        # TODO (savin): Enable lazy instantiation of package
        ctx.obj.package = None

    if ctx.invoked_subcommand is None:
        ctx.invoke(check)


def _check(echo, graph, flow, environment, pylint=True, warnings=False, **kwargs):
    echo("Validating your flow...", fg="magenta", bold=False)
    linter = lint.linter
    # TODO set linter settings
    linter.run_checks(graph, **kwargs)
    echo("The graph looks good!", fg="green", bold=True, indent=True)
    if pylint:
        echo("Running pylint...", fg="magenta", bold=False)
        fname = inspect.getfile(flow.__class__)
        pylint = PyLint(fname)
        if pylint.has_pylint():
            pylint_is_happy, pylint_exception_msg = pylint.run(
                warnings=warnings,
                pylint_config=environment.pylint_config(),
                logger=echo_always,
            )

            if pylint_is_happy:
                echo("Pylint is happy!", fg="green", bold=True, indent=True)
            else:
                echo(
                    "Pylint couldn't analyze your code.\n\tPylint exception: %s"
                    % pylint_exception_msg,
                    fg="red",
                    bold=True,
                    indent=True,
                )
                echo("Skipping Pylint checks.", fg="red", bold=True, indent=True)
        else:
            echo(
                "Pylint not found, so extra checks are disabled.",
                fg="green",
                indent=True,
                bold=False,
            )


def print_metaflow_exception(ex):
    echo_always(ex.headline, indent=True, nl=False, bold=True)
    location = ""
    if ex.source_file is not None:
        location += " in file %s" % ex.source_file
    if ex.line_no is not None:
        location += " on line %d" % ex.line_no
    location += ":"
    echo_always(location, bold=True)
    echo_always(ex.message, indent=True, bold=False, padding_bottom=True)


def print_unknown_exception(ex):
    echo_always("Internal error", indent=True, bold=True)
    echo_always(traceback.format_exc(), highlight=None, highlight_bold=False)


class CliState(object):
    def __init__(self, flow):
        self.flow = flow


def main(flow, args=None, handle_exceptions=True, entrypoint=None):
    # Ignore warning(s) and prevent spamming the end-user.
    # TODO: This serves as a short term workaround for RuntimeWarning(s) thrown
    # in py3.8 related to log buffering (bufsize=1).
    import warnings

    warnings.filterwarnings("ignore")
    if entrypoint is None:
        entrypoint = [sys.executable, sys.argv[0]]

    state = CliState(flow)
    state.entrypoint = entrypoint

    try:
        if args is None:
            start(auto_envvar_prefix="METAFLOW", obj=state)
        else:
            try:
                start(args=args, obj=state, auto_envvar_prefix="METAFLOW")
            except SystemExit as e:
                return e.code
    except MetaflowException as x:
        if handle_exceptions:
            print_metaflow_exception(x)
            sys.exit(1)
        else:
            raise
    except Exception as x:
        if handle_exceptions:
            print_unknown_exception(x)
            sys.exit(1)
        else:
            raise
    finally:
        if hasattr(state, "monitor") and state.monitor is not None:
            state.monitor.terminate()
        if hasattr(state, "event_logger") and state.event_logger is not None:
            state.event_logger.terminate()



import functools
from html.parser import HTMLParser


class MyBaseException(Exception):
    pass


class SomeException(MyBaseException):
    pass


class ExceptionAndClass(MyBaseException):
    def __init__(self, *args):
        super().__init__(*args)

    def method_on_exception(self):
        return "method_on_exception"

    def __str__(self):
        return "ExceptionAndClass Str: %s" % super().__str__()


class ExceptionAndClassChild(ExceptionAndClass):
    def __init__(self, *args):
        super().__init__(*args)

    def method_on_child_exception(self):
        return "method_on_child_exception"

    def __str__(self):
        return "ExceptionAndClassChild Str: %s" % super().__str__()


class BaseClass(HTMLParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._output = []

    def handle_starttag(self, tag, attrs):
        self._output.append(tag)
        return super().handle_starttag(tag, attrs)

    def get_output(self):
        return self._output


class ChildClass(BaseClass):
    def handle_endtag(self, tag):
        self._output.append(tag)
        return super().handle_endtag(tag)


class TestClass1(object):
    cls_object = 25

    def __init__(self, value):
        self._value = value
        self._value2 = 123

    def unsupported_method(self):
        pass

    def print_value(self):
        return self._value

    def __str__(self):
        return "My str representation is %s" % str(self._value)

    def __repr__(self):
        return "My repr representation is %s" % str(self._value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def to_class2(self, count, stride=1):
        return TestClass2(self._value, stride, count)

    @staticmethod
    def static_method(val):
        return val + 42

    @classmethod
    def class_method(cls):
        return cls.cls_object

    @property
    def override_value(self):
        return self._value2

    @override_value.setter
    def override_value(self, value):
        self._value2 = value

    def __hidden(self, name, value):
        setattr(self, name, value)

    def weird_indirection(self, name):
        return functools.partial(self.__hidden, name)

    def returnChild(self):
        return ChildClass()

    def raiseOrReturnValueError(self, doRaise=False):
        if doRaise:
            raise ValueError("I raised")
        return ValueError("I returned")

    def raiseOrReturnSomeException(self, doRaise=False):
        if doRaise:
            raise SomeException("I raised")
        return SomeException("I returned")

    def raiseOrReturnExceptionAndClass(self, doRaise=False):
        if doRaise:
            raise ExceptionAndClass("I raised")
        return ExceptionAndClass("I returned")

    def raiseOrReturnExceptionAndClassChild(self, doRaise=False):
        if doRaise:
            raise ExceptionAndClassChild("I raised")
        return ExceptionAndClassChild("I returned")


class TestClass2(object):
    def __init__(self, value, stride, count):
        self._mylist = [value + stride * i for i in range(count)]

    def something(self, val):
        return "Test2:Something:%s" % val

    def __iter__(self):
        self._pos = 0
        return self

    def __next__(self):
        if self._pos < len(self._mylist):
            self._pos += 1
            return self._mylist[self._pos - 1]
        raise StopIteration


def test_func(*args, **kwargs):
    return "In test func"


test_value = 1



from metaflow._vendor import click
from datetime import datetime, timezone
from metaflow.tagging_util import validate_tags
from metaflow.metadata_provider import MetaDatum


@click.group()
def cli():
    pass


@cli.group(help="Commands related to spot metadata.")
def spot_metadata():
    pass


@spot_metadata.command(help="Record spot termination metadata for a task.")
@click.option(
    "--run-id",
    required=True,
    help="Run ID for which metadata is to be recorded.",
)
@click.option(
    "--step-name",
    required=True,
    help="Step Name for which metadata is to be recorded.",
)
@click.option(
    "--task-id",
    required=True,
    help="Task ID for which metadata is to be recorded.",
)
@click.option(
    "--termination-notice-time",
    required=True,
    help="Spot termination notice time.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    required=False,
    default=None,
    help="List of tags.",
)
@click.pass_obj
def record(obj, run_id, step_name, task_id, termination_notice_time, tags=None):
    validate_tags(tags)

    tag_list = list(tags) if tags else []

    entries = [
        MetaDatum(
            field="spot-termination-received-at",
            value=datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
            type="spot-termination-received-at",
            tags=tag_list,
        ),
        MetaDatum(
            field="spot-termination-time",
            value=termination_notice_time,
            type="spot-termination-time",
            tags=tag_list,
        ),
    ]

    obj.metadata.register_metadata(
        run_id=run_id, step_name=step_name, task_id=task_id, metadata=entries
    )



import sys
from metaflow.metaflow_config import (
    OTEL_ENDPOINT,
    ZIPKIN_ENDPOINT,
    CONSOLE_TRACE_ENABLED,
    DISABLE_TRACING,
    DEBUG_TRACING,
)
from functools import wraps
import contextlib
from typing import Dict


def init_tracing():
    pass


@contextlib.contextmanager
def post_fork():
    yield


def cli(name: str):
    def cli_wrap(func):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper_func

    return cli_wrap


def inject_tracing_vars(env_dict: Dict[str, str]) -> Dict[str, str]:
    return env_dict


def get_trace_id() -> str:
    return ""


@contextlib.contextmanager
def traced(name, attrs=None):
    if attrs is None:
        attrs = {}
    yield


def tracing(func):
    @wraps(func)
    def wrapper_func(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper_func


if not DISABLE_TRACING and (CONSOLE_TRACE_ENABLED or OTEL_ENDPOINT or ZIPKIN_ENDPOINT):
    try:
        from .tracing_modules import (
            init_tracing,
            post_fork,
            cli,
            inject_tracing_vars,
            get_trace_id,
            traced,
            tracing,
        )
    except ImportError as e:
        # We keep the errors silent by default so that having tracing environment variables present
        # does not affect users with no need for tracing.
        if DEBUG_TRACING:
            print(e.msg, file=sys.stderr)



import os
import random
import time

import requests

from typing import List
from metaflow.exception import (
    MetaflowException,
    MetaflowInternalError,
    MetaflowTaggingError,
)
from metaflow.metadata_provider import MetadataProvider
from metaflow.metadata_provider.heartbeat import HB_URL_KEY
from metaflow.metaflow_config import SERVICE_HEADERS, SERVICE_RETRY_COUNT, SERVICE_URL
from metaflow.sidecar import Message, MessageTypes, Sidecar
from urllib.parse import urlencode
from metaflow.util import version_parse


# Define message enums
class HeartbeatTypes(object):
    RUN = 1
    TASK = 2


class ServiceException(MetaflowException):
    headline = "Metaflow service error"

    def __init__(self, msg, http_code=None, body=None):
        self.http_code = None if http_code is None else int(http_code)
        self.response = body
        super(ServiceException, self).__init__(msg)


class ServiceMetadataProvider(MetadataProvider):
    TYPE = "service"

    _session = requests.Session()
    _session.mount(
        "http://",
        requests.adapters.HTTPAdapter(
            pool_connections=20,
            pool_maxsize=20,
            max_retries=0,  # Handle retries explicitly
            pool_block=False,
        ),
    )
    _session.mount(
        "https://",
        requests.adapters.HTTPAdapter(
            pool_connections=20, pool_maxsize=20, max_retries=0, pool_block=False
        ),
    )

    _supports_attempt_gets = None
    _supports_tag_mutation = None

    def __init__(self, environment, flow, event_logger, monitor):
        super(ServiceMetadataProvider, self).__init__(
            environment, flow, event_logger, monitor
        )
        self.url_task_template = os.path.join(
            SERVICE_URL,
            "flows/{flow_id}/runs/{run_number}/steps/{step_name}/tasks/{task_id}/heartbeat",
        )
        self.url_run_template = os.path.join(
            SERVICE_URL, "flows/{flow_id}/runs/{run_number}/heartbeat"
        )
        self.sidecar = None

    @classmethod
    def compute_info(cls, val):
        v = val.rstrip("/")
        for i in range(SERVICE_RETRY_COUNT):
            try:
                resp = cls._session.get(
                    os.path.join(v, "ping"), headers=SERVICE_HEADERS.copy()
                )
                resp.raise_for_status()
            except:  # noqa E722
                time.sleep(2 ** (i - 1))
            else:
                return v

        raise ValueError("Metaflow service [%s] unreachable." % v)

    @classmethod
    def default_info(cls):
        return SERVICE_URL

    def version(self):
        return self._version(self._monitor)

    def new_run_id(self, tags=None, sys_tags=None):
        v, _ = self._new_run(tags=tags, sys_tags=sys_tags)
        return v

    def register_run_id(self, run_id, tags=None, sys_tags=None):
        try:
            # don't try to register an integer ID which was obtained
            # from the metadata service in the first place
            int(run_id)
            return False
        except ValueError:
            _, did_create = self._new_run(run_id, tags=tags, sys_tags=sys_tags)
            return did_create

    def new_task_id(self, run_id, step_name, tags=None, sys_tags=None):
        v, _ = self._new_task(run_id, step_name, tags=tags, sys_tags=sys_tags)
        return v

    def register_task_id(
        self, run_id, step_name, task_id, attempt=0, tags=None, sys_tags=None
    ):
        try:
            # don't try to register an integer ID which was obtained
            # from the metadata service in the first place
            int(task_id)
        except ValueError:
            _, did_create = self._new_task(
                run_id,
                step_name,
                task_id=task_id,
                attempt=attempt,
                tags=tags,
                sys_tags=sys_tags,
            )
            return did_create
        else:
            self._register_system_metadata(run_id, step_name, task_id, attempt)
            return False

    def _start_heartbeat(
        self, heartbeat_type, flow_id, run_id, step_name=None, task_id=None
    ):
        if self._already_started():
            # A single ServiceMetadataProvider instance can not start
            # multiple heartbeat side cars of any type/combination. Either a
            # single run heartbeat or a single task heartbeat can be started
            raise Exception("heartbeat already started")
        # create init message
        payload = {}
        if heartbeat_type == HeartbeatTypes.TASK:
            # create task heartbeat
            data = {
                "flow_id": flow_id,
                "run_number": run_id,
                "step_name": step_name,
                "task_id": task_id,
            }
            payload[HB_URL_KEY] = self.url_task_template.format(**data)
        elif heartbeat_type == HeartbeatTypes.RUN:
            # create run heartbeat
            data = {"flow_id": flow_id, "run_number": run_id}

            payload[HB_URL_KEY] = self.url_run_template.format(**data)
        else:
            raise Exception("invalid heartbeat type")
        service_version = self.version()
        payload["service_version"] = service_version
        # start sidecar
        if service_version is None or version_parse(service_version) < version_parse(
            "2.0.4"
        ):
            # if old version of the service is running
            # then avoid running real heartbeat sidecar process
            self.sidecar = Sidecar("none")
        else:
            self.sidecar = Sidecar("heartbeat")
        self.sidecar.start()
        self.sidecar.send(Message(MessageTypes.BEST_EFFORT, payload))

    def start_run_heartbeat(self, flow_id, run_id):
        self._start_heartbeat(HeartbeatTypes.RUN, flow_id, run_id)

    def start_task_heartbeat(self, flow_id, run_id, step_name, task_id):
        self._start_heartbeat(HeartbeatTypes.TASK, flow_id, run_id, step_name, task_id)

    def _already_started(self):
        return self.sidecar is not None

    def stop_heartbeat(self):
        self.sidecar.terminate()

    def register_data_artifacts(
        self, run_id, step_name, task_id, attempt_id, artifacts
    ):
        url = ServiceMetadataProvider._obj_path(
            self._flow_name, run_id, step_name, task_id
        )
        url += "/artifact"
        data = self._artifacts_to_json(
            run_id, step_name, task_id, attempt_id, artifacts
        )
        self._request(self._monitor, url, "POST", data)

    def register_metadata(self, run_id, step_name, task_id, metadata):
        url = ServiceMetadataProvider._obj_path(
            self._flow_name, run_id, step_name, task_id
        )
        url += "/metadata"
        data = self._metadata_to_json(run_id, step_name, task_id, metadata)
        self._request(self._monitor, url, "POST", data)

    @classmethod
    def _mutate_user_tags_for_run(
        cls, flow_id, run_id, tags_to_add=None, tags_to_remove=None
    ):
        min_service_version_with_tag_mutation = "2.3.0"
        if cls._supports_tag_mutation is None:
            version = cls._version(None)
            cls._supports_tag_mutation = version is not None and version_parse(
                version
            ) >= version_parse(min_service_version_with_tag_mutation)
        if not cls._supports_tag_mutation:
            raise ServiceException(
                "Adding or removing tags on a run requires the Metaflow service to be "
                "at least version %s. Please upgrade your service."
                % (min_service_version_with_tag_mutation,)
            )

        url = ServiceMetadataProvider._obj_path(flow_id, run_id) + "/tag/mutate"
        tag_mutation_data = {
            # mutate_user_tags_for_run() should have already ensured that this is a list, so let's be tolerant here
            "tags_to_add": list(tags_to_add or []),
            "tags_to_remove": list(tags_to_remove or []),
        }
        tries = 1
        status_codes_seen = set()
        # try up to 10 times, with a gentle exponential backoff (1.4-1.6x)
        while True:
            resp, _ = cls._request(
                None, url, "PATCH", data=tag_mutation_data, return_raw_resp=True
            )
            status_codes_seen.add(resp.status_code)
            # happy path
            if resp.status_code < 300:
                return frozenset(resp.json()["tags"])
            # definitely NOT retriable
            if resp.status_code in (400, 422):
                raise MetaflowTaggingError("Metadata service says: %s" % (resp.text,))
            # if we get here, mutation failure is possibly retriable
            if tries >= 10:
                # if we ever received 409 on any of our attempts, report "conflicting updates" blurb to user
                if 409 in status_codes_seen:
                    raise MetaflowTaggingError(
                        "Tagging failed due to too many conflicting updates from other processes"
                    )
                # No 409's seen... raise a more generic error
                raise MetaflowTaggingError("Tagging failed after %d tries" % tries)
            time.sleep(0.3 * random.uniform(1.4, 1.6) ** tries)
            tries += 1

    @classmethod
    def _get_object_internal(
        cls, obj_type, obj_order, sub_type, sub_order, filters, attempt, *args
    ):
        if attempt is not None:
            if cls._supports_attempt_gets is None:
                version = cls._version(None)
                cls._supports_attempt_gets = version is not None and version_parse(
                    version
                ) >= version_parse("2.0.6")
            if not cls._supports_attempt_gets:
                raise ServiceException(
                    "Getting specific attempts of Tasks or Artifacts requires "
                    "the metaflow service to be at least version 2.0.6. Please "
                    "upgrade your service"
                )

        if sub_type == "self":
            if obj_type == "artifact":
                # Special case with the artifacts; we add the attempt
                url = ServiceMetadataProvider._obj_path(
                    *args[:obj_order], attempt=attempt
                )
            else:
                url = ServiceMetadataProvider._obj_path(*args[:obj_order])
            try:
                v, _ = cls._request(None, url, "GET")
                return MetadataProvider._apply_filter([v], filters)[0]
            except ServiceException as ex:
                if ex.http_code == 404:
                    return None
                raise

        # For the other types, we locate all the objects we need to find and return them
        if obj_type != "root":
            url = ServiceMetadataProvider._obj_path(*args[:obj_order])
        else:
            url = ""
        if sub_type == "metadata":
            url += "/metadata"
        elif sub_type == "artifact" and obj_type == "task" and attempt is not None:
            url += "/attempt/%s/artifacts" % attempt
        else:
            url += "/%ss" % sub_type
        try:
            v, _ = cls._request(None, url, "GET")
            return MetadataProvider._apply_filter(v, filters)
        except ServiceException as ex:
            if ex.http_code == 404:
                return None
            raise

    def _new_run(self, run_id=None, tags=None, sys_tags=None):
        # first ensure that the flow exists
        self._get_or_create("flow")
        run, did_create = self._get_or_create(
            "run", run_id, tags=tags, sys_tags=sys_tags
        )
        return str(run["run_number"]), did_create

    def _new_task(
        self, run_id, step_name, task_id=None, attempt=0, tags=None, sys_tags=None
    ):
        # first ensure that the step exists
        self._get_or_create("step", run_id, step_name)
        task, did_create = self._get_or_create(
            "task", run_id, step_name, task_id, tags=tags, sys_tags=sys_tags
        )
        if did_create:
            self._register_system_metadata(run_id, step_name, task["task_id"], attempt)
        return task["task_id"], did_create

    @classmethod
    def filter_tasks_by_metadata(
        cls,
        flow_name: str,
        run_id: str,
        step_name: str,
        field_name: str,
        pattern: str,
    ) -> List[str]:
        """
        Filter tasks by metadata field and pattern, returning task pathspecs that match criteria.

        Parameters
        ----------
        flow_name : str
            Flow name, that the run belongs to.
        run_id: str
            Run id, together with flow_id, that identifies the specific Run whose tasks to query
        step_name: str
            Step name to query tasks from
        field_name: str
            Metadata field name to query
        pattern: str
            Pattern to match in metadata field value

        Returns
        -------
        List[str]
            List of task pathspecs that satisfy the query
        """
        query_params = {}

        if pattern == ".*":
            # we do not need to filter tasks at all if pattern allows 'any'
            query_params = {}
        else:
            if field_name:
                query_params["metadata_field_name"] = field_name
            if pattern:
                query_params["pattern"] = pattern

        url = ServiceMetadataProvider._obj_path(flow_name, run_id, step_name)
        url = f"{url}/filtered_tasks?{urlencode(query_params)}"

        try:
            resp, _ = cls._request(None, url, "GET")
        except Exception as e:
            if e.http_code == 404:
                # filter_tasks_by_metadata endpoint does not exist in the version of metadata service
                # deployed currently. Raise a more informative error message.
                raise MetaflowInternalError(
                    "The version of metadata service deployed currently does not support filtering tasks by metadata. "
                    "Upgrade Metadata service to version 2.5.0 or greater to use this feature."
                ) from e
            # Other unknown exception
            raise e
        return resp

    @staticmethod
    def _obj_path(
        flow_name,
        run_id=None,
        step_name=None,
        task_id=None,
        artifact_name=None,
        attempt=None,
    ):
        object_path = "/flows/%s" % flow_name
        if run_id is not None:
            object_path += "/runs/%s" % run_id
        if step_name is not None:
            object_path += "/steps/%s" % step_name
        if task_id is not None:
            object_path += "/tasks/%s" % task_id
        if artifact_name is not None:
            object_path += "/artifacts/%s" % artifact_name
        if attempt is not None:
            object_path += "/attempt/%s" % attempt
        return object_path

    @staticmethod
    def _create_path(obj_type, flow_name, run_id=None, step_name=None):
        create_path = "/flows/%s" % flow_name
        if obj_type == "flow":
            return create_path
        if obj_type == "run":
            return create_path + "/run"
        create_path += "/runs/%s/steps/%s" % (run_id, step_name)
        if obj_type == "step":
            return create_path + "/step"
        return create_path + "/task"

    def _get_or_create(
        self,
        obj_type,
        run_id=None,
        step_name=None,
        task_id=None,
        tags=None,
        sys_tags=None,
    ):
        if tags is None:
            tags = set()
        if sys_tags is None:
            sys_tags = set()

        def create_object():
            data = self._object_to_json(
                obj_type,
                run_id,
                step_name,
                task_id,
                self.sticky_tags.union(tags),
                self.sticky_sys_tags.union(sys_tags),
            )
            return self._request(
                self._monitor, create_path, "POST", data=data, retry_409_path=obj_path
            )

        always_create = False
        obj_path = self._obj_path(self._flow_name, run_id, step_name, task_id)
        create_path = self._create_path(obj_type, self._flow_name, run_id, step_name)
        if obj_type == "run" and run_id is None:
            always_create = True
        elif obj_type == "task" and task_id is None:
            always_create = True

        if always_create:
            return create_object()

        try:
            return self._request(self._monitor, obj_path, "GET")
        except ServiceException as ex:
            if ex.http_code == 404:
                return create_object()
            else:
                raise

    # TODO _request() needs a more deliberate refactor at some point, it looks quite overgrown.
    @classmethod
    def _request(
        cls,
        monitor,
        path,
        method,
        data=None,
        retry_409_path=None,
        return_raw_resp=False,
    ):
        if cls.INFO is None:
            raise MetaflowException(
                "Missing Metaflow Service URL. "
                "Specify with METAFLOW_SERVICE_URL environment variable"
            )
        supported_methods = ("GET", "PATCH", "POST")
        if method not in supported_methods:
            raise MetaflowException(
                "Only these methods are supported: %s, but got %s"
                % (supported_methods, method)
            )
        url = os.path.join(cls.INFO, path.lstrip("/"))
        for i in range(SERVICE_RETRY_COUNT):
            try:
                if method == "GET":
                    if monitor:
                        with monitor.measure("metaflow.service_metadata.get"):
                            resp = cls._session.get(url, headers=SERVICE_HEADERS.copy())
                    else:
                        resp = cls._session.get(url, headers=SERVICE_HEADERS.copy())
                elif method == "POST":
                    if monitor:
                        with monitor.measure("metaflow.service_metadata.post"):
                            resp = cls._session.post(
                                url, headers=SERVICE_HEADERS.copy(), json=data
                            )
                    else:
                        resp = cls._session.post(
                            url, headers=SERVICE_HEADERS.copy(), json=data
                        )
                elif method == "PATCH":
                    if monitor:
                        with monitor.measure("metaflow.service_metadata.patch"):
                            resp = cls._session.patch(
                                url, headers=SERVICE_HEADERS.copy(), json=data
                            )
                    else:
                        resp = cls._session.patch(
                            url, headers=SERVICE_HEADERS.copy(), json=data
                        )
                else:
                    raise MetaflowInternalError("Unexpected HTTP method %s" % (method,))
            except MetaflowInternalError:
                raise
            except:  # noqa E722
                if monitor:
                    with monitor.count("metaflow.service_metadata.failed_request"):
                        if i == SERVICE_RETRY_COUNT - 1:
                            raise
                else:
                    if i == SERVICE_RETRY_COUNT - 1:
                        raise
                resp = None
            else:
                if return_raw_resp:
                    return resp, True
                if resp.status_code < 300:
                    return resp.json(), True
                elif resp.status_code == 409 and data is not None:
                    # a special case: the post fails due to a conflict
                    # this could occur when we missed a success response
                    # from the first POST request but the request
                    # actually went though, so a subsequent POST
                    # returns 409 (conflict) or we end up with a
                    # conflict while running on AWS Step Functions
                    # instead of retrying the post we retry with a get since
                    # the record is guaranteed to exist
                    if retry_409_path:
                        v, _ = cls._request(monitor, retry_409_path, "GET")
                        return v, False
                    else:
                        return None, False
                elif resp.status_code != 503:
                    raise ServiceException(
                        "Metadata request (%s) failed (code %s): %s"
                        % (path, resp.status_code, resp.text),
                        resp.status_code,
                        resp.text,
                    )
            time.sleep(2**i)
        if resp:
            raise ServiceException(
                "Metadata request (%s) failed (code %s): %s"
                % (path, resp.status_code, resp.text),
                resp.status_code,
                resp.text,
            )
        else:
            raise ServiceException("Metadata request (%s) failed" % path)

    @classmethod
    def _version(cls, monitor):
        if cls.INFO is None:
            raise MetaflowException(
                "Missing Metaflow Service URL. "
                "Specify with METAFLOW_SERVICE_URL environment variable"
            )
        path = "ping"
        url = os.path.join(cls.INFO, path)
        for i in range(SERVICE_RETRY_COUNT):
            try:
                if monitor:
                    with monitor.measure("metaflow.service_metadata.get"):
                        resp = cls._session.get(url, headers=SERVICE_HEADERS.copy())
                else:
                    resp = cls._session.get(url, headers=SERVICE_HEADERS.copy())
            except:
                if monitor:
                    with monitor.count("metaflow.service_metadata.failed_request"):
                        if i == SERVICE_RETRY_COUNT - 1:
                            raise
                else:
                    if i == SERVICE_RETRY_COUNT - 1:
                        raise
                resp = None
            else:
                if resp.status_code < 300:
                    return resp.headers.get("METADATA_SERVICE_VERSION", None)
                elif resp.status_code not in (503, 500):
                    raise ServiceException(
                        "Metadata request (%s) failed"
                        " (code %s): %s" % (url, resp.status_code, resp.text),
                        resp.status_code,
                        resp.text,
                    )
            time.sleep(2**i)
        if resp:
            raise ServiceException(
                "Metadata request (%s) failed (code %s): %s"
                % (url, resp.status_code, resp.text),
                resp.status_code,
                resp.text,
            )
        else:
            raise ServiceException("Metadata request (%s) failed" % url)



import importlib
import json
import re

from functools import partial
from typing import Any, Callable, Dict, List, NewType, Tuple, TypeVar, Union, overload

from .flowspec import FlowSpec, _FlowState
from .exception import (
    MetaflowInternalError,
    MetaflowException,
    InvalidDecoratorAttribute,
)

from .debug import debug
from .parameters import current_flow
from .user_configs.config_parameters import (
    UNPACK_KEY,
    resolve_delayed_evaluator,
    unpack_delayed_evaluator,
)
from .user_decorators.mutable_flow import MutableFlow
from .user_decorators.mutable_step import MutableStep
from .user_decorators.user_flow_decorator import FlowMutator, FlowMutatorMeta
from .user_decorators.user_step_decorator import (
    StepMutator,
    UserStepDecoratorBase,
    UserStepDecoratorMeta,
)
from .metaflow_config import SPIN_ALLOWED_DECORATORS
from metaflow._vendor import click


class BadStepDecoratorException(MetaflowException):
    headline = "Syntax error"

    def __init__(self, deco, func):
        msg = (
            "You tried to apply decorator '{deco}' on '{func}' which is "
            "not declared as a @step. Make sure you apply this decorator "
            "on a function which has @step on the line just before the "
            "function name and @{deco} is above @step.".format(
                deco=deco, func=getattr(func, "__name__", str(func))
            )
        )
        super(BadStepDecoratorException, self).__init__(msg)


class BadFlowDecoratorException(MetaflowException):
    headline = "Syntax error"

    def __init__(self, deconame):
        msg = (
            "Decorator '%s' can be applied only to FlowSpecs. Make sure "
            "the decorator is above a class definition." % deconame
        )
        super(BadFlowDecoratorException, self).__init__(msg)


class UnknownStepDecoratorException(MetaflowException):
    headline = "Unknown step decorator"

    def __init__(self, deconame):
        decos = ", ".join(
            [
                x
                for x in UserStepDecoratorMeta.all_decorators().keys()
                if not x.endswith("_internal")
            ]
        )

        msg = (
            "Unknown step decorator *{deconame}*. The following decorators are "
            "supported: *{decos}*".format(deconame=deconame, decos=decos)
        )
        super(UnknownStepDecoratorException, self).__init__(msg)


class DuplicateStepDecoratorException(MetaflowException):
    headline = "Duplicate decorators"

    def __init__(self, deco, func):
        msg = (
            "Step '{step}' already has a decorator '{deco}'. "
            "You can specify this decorator only once.".format(
                step=func.__name__, deco=deco
            )
        )
        super(DuplicateStepDecoratorException, self).__init__(msg)


class UnknownFlowDecoratorException(MetaflowException):
    headline = "Unknown flow decorator"

    def __init__(self, deconame):
        decos = ", ".join(FlowMutatorMeta.all_decorators().keys())
        msg = (
            "Unknown flow decorator *{deconame}*. The following decorators are "
            "supported: *{decos}*".format(deconame=deconame, decos=decos)
        )
        super(UnknownFlowDecoratorException, self).__init__(msg)


class DuplicateFlowDecoratorException(MetaflowException):
    headline = "Duplicate decorators"

    def __init__(self, deco):
        msg = (
            "Flow already has a decorator '{deco}'. "
            "You can specify each decorator only once.".format(deco=deco)
        )
        super(DuplicateFlowDecoratorException, self).__init__(msg)


class Decorator(object):
    """
    Base class for all decorators.
    """

    name = "NONAME"
    defaults = {}
    # `allow_multiple` allows setting many decorators of the same type to a step/flow.
    allow_multiple = False

    def __init__(self, attributes=None, statically_defined=False, inserted_by=None):
        self.attributes = self.defaults.copy()
        self.statically_defined = statically_defined
        self.inserted_by = inserted_by
        self._user_defined_attributes = set()
        self._ran_init = False

        if attributes:
            for k, v in attributes.items():
                if k in self.defaults or k.startswith(UNPACK_KEY):
                    self.attributes[k] = v
                    if not k.startswith(UNPACK_KEY):
                        self._user_defined_attributes.add(k)
                else:
                    raise InvalidDecoratorAttribute(self.name, k, self.defaults)

    def init(self):
        """
        Initializes the decorator. In general, any operation you would do in __init__
        should be done here.
        """
        pass

    def external_init(self):
        # In some cases (specifically when using remove_decorator), we may need to call
        # init multiple times. Short-circuit re-evaluating.
        if self._ran_init:
            return

        # Note that by design, later values override previous ones.
        self.attributes, new_user_attributes = unpack_delayed_evaluator(self.attributes)
        self._user_defined_attributes.update(new_user_attributes)
        self.attributes = resolve_delayed_evaluator(self.attributes, to_dict=True)

        if "init" in self.__class__.__dict__:
            self.init()
        self._ran_init = True

    @classmethod
    def extract_args_kwargs_from_decorator_spec(cls, deco_spec):
        if len(deco_spec) == 0:
            return [], {}

        attrs = {}
        # TODO: Do we really want to allow spaces in the names of attributes?!?
        for a in re.split(r""",(?=[\s\w]+=)""", deco_spec):
            name, val = a.split("=", 1)
            try:
                val_parsed = json.loads(val.strip().replace('\\"', '"'))
            except json.JSONDecodeError:
                # In this case, we try to convert to either an int or a float or
                # leave as is. Prefer ints if possible.
                try:
                    val_parsed = int(val.strip())
                except ValueError:
                    try:
                        val_parsed = float(val.strip())
                    except ValueError:
                        val_parsed = val.strip()

            attrs[name.strip()] = val_parsed

        return [], attrs

    @classmethod
    def parse_decorator_spec(cls, deco_spec):
        if len(deco_spec) == 0:
            return cls()

        _, kwargs = cls.extract_args_kwargs_from_decorator_spec(deco_spec)
        return cls(attributes=kwargs)

    def make_decorator_spec(self):
        # Make sure all attributes are evaluated
        self.external_init()
        attrs = {k: v for k, v in self.attributes.items() if v is not None}
        if attrs:
            attr_list = []
            # We dump simple types directly as string to get around the nightmare quote
            # escaping but for more complex types (typically dictionaries or lists),
            # we dump using JSON.
            for k, v in attrs.items():
                if isinstance(v, (int, float, str)):
                    attr_list.append("%s=%s" % (k, str(v)))
                else:
                    attr_list.append("%s=%s" % (k, json.dumps(v).replace('"', '\\"')))

            attrstr = ",".join(attr_list)
            return "%s:%s" % (self.name, attrstr)
        else:
            return self.name

    def get_args_kwargs(self) -> Tuple[List[Any], Dict[str, Any]]:
        """
        Get the arguments and keyword arguments of the decorator.

        Returns
        -------
        Tuple[List[Any], Dict[str, Any]]
            A tuple containing a list of arguments and a dictionary of keyword arguments.
        """
        return [], dict(self.attributes)

    def __str__(self):
        mode = "static" if self.statically_defined else "dynamic"
        if self.inserted_by:
            mode += " (inserted by %s)" % " from ".join(self.inserted_by)
        attrs = " ".join("%s=%s" % x for x in self.attributes.items())
        if attrs:
            attrs = " " + attrs
        fmt = "%s<%s%s>" % (self.name, mode, attrs)
        return fmt


class FlowDecorator(Decorator):
    options = {}

    def __init__(self, *args, **kwargs):
        super(FlowDecorator, self).__init__(*args, **kwargs)

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        """
        Called when all decorators have been created for this flow.
        """
        pass

    def get_top_level_options(self):
        """
        Return a list of option-value pairs that correspond to top-level
        options that should be passed to subprocesses (tasks). The option
        names should be a subset of the keys in self.options.

        If the decorator has a non-empty set of options in `self.options`, you
        probably want to return the assigned values in this method.
        """
        return []


# compare this to parameters.add_custom_parameters
def add_decorator_options(cmd):
    flow_cls = getattr(current_flow, "flow_cls", None)
    if flow_cls is None:
        return cmd

    seen = {}
    existing_params = set(p.name.lower() for p in cmd.params)
    # Add decorator options
    for deco in flow_decorators(flow_cls):
        for option, kwargs in deco.options.items():
            if option in seen:
                msg = (
                    "Flow decorator '%s' uses an option '%s' which is also "
                    "used by the decorator '%s'. This is a bug in Metaflow. "
                    "Please file a ticket on GitHub."
                    % (deco.name, option, seen[option])
                )
                raise MetaflowInternalError(msg)
            elif deco.name.lower() in existing_params:
                raise MetaflowInternalError(
                    "Flow decorator '%s' uses an option '%s' which is a reserved "
                    "keyword. Please use a different option name." % (deco.name, option)
                )
            else:
                kwargs["envvar"] = "METAFLOW_FLOW_%s" % option.upper()
                seen[option] = deco.name
                cmd.params.insert(0, click.Option(("--" + option,), **kwargs))
    return cmd


def flow_decorators(flow_cls):
    return [d for deco_list in flow_cls._flow_decorators.values() for d in deco_list]


class StepDecorator(Decorator):
    """
    Base class for all step decorators.

    Example:

    @my_decorator
    @step
    def a(self):
        pass

    @my_decorator
    @step
    def b(self):
        pass

    To make the above work, define a subclass

    class MyDecorator(StepDecorator):
        name = "my_decorator"

    and include it in plugins.STEP_DECORATORS. Now both a() and b()
    get an instance of MyDecorator, so you can keep step-specific
    state easily.

    TODO (savin): Initialize the decorators with flow, graph,
                  step.__name__ etc., so that we don't have to
                  pass them around with every lifecycle call.
    """

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        """
        Called when all decorators have been created for this step
        """
        pass

    def package_init(self, flow, step_name, environment):
        """
        Called to determine package components
        """
        pass

    def add_to_package(self):
        """
        Called to add custom files needed for this environment. This hook will be
        called in the `MetaflowPackage` class where metaflow compiles the code package
        tarball. This hook can return one of two things (the first is for backwards
        compatibility -- move to the second):
          - a generator yielding a tuple of `(file_path, arcname)` to add files to
            the code package. `file_path` is the path to the file on the local filesystem
            and `arcname` is the path relative to the packaged code.
          - a generator yielding a tuple of `(content, arcname, type)` where:
            - type is one of
            ContentType.{USER_CONTENT, CODE_CONTENT, MODULE_CONTENT, OTHER_CONTENT}
            - for USER_CONTENT:
              - the file will be included relative to the directory containing the
                user's flow file.
              - content: path to the file to include
              - arcname: path relative to the directory containing the user's flow file
            - for CODE_CONTENT:
              - the file will be included relative to the code directory in the package.
                This will be the directory containing `metaflow`.
              - content: path to the file to include
              - arcname: path relative to the code directory in the package
            - for MODULE_CONTENT:
              - the module will be added to the code package as a python module. It will
                be accessible as usual (import <module_name>)
              - content: name of the module
              - arcname: None (ignored)
            - for OTHER_CONTENT:
              - the file will be included relative to any other configuration/metadata
                files for the flow
              - content: path to the file to include
              - arcname: path relative to the config directory in the package
        """
        return []

    def step_task_retry_count(self):
        """
        Called to determine the number of times this task should be retried.
        Returns a tuple of (user_code_retries, error_retries). Error retries
        are attempts to run the process after the user code has failed all
        its retries.

        Typically, the runtime takes the maximum of retry counts across
        decorators and user specification to determine the task retry count.
        If you want to force no retries, return the special values (None, None).
        """
        return 0, 0

    def runtime_init(self, flow, graph, package, run_id):
        """
        Top-level initialization before anything gets run in the runtime
        context.
        """
        pass

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        """
        Called when the runtime has created a task related to this step.
        """
        pass

    def runtime_finished(self, exception):
        """
        Called when the runtime created task finishes or encounters an interrupt/exception.
        """
        pass

    def runtime_step_cli(
        self, cli_args, retry_count, max_user_code_retries, ubf_context
    ):
        """
        Access the command line for a step execution in the runtime context.
        """
        pass

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        """
        Run before the step function in the task context.
        """
        pass

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        return step_func

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        """
        Run after the step function has finished successfully in the task
        context.
        """
        pass

    def task_exception(
        self, exception, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        """
        Run if the step function raised an exception in the task context.

        If this method returns True, it is assumed that the exception has
        been taken care of and the flow may continue.
        """
        pass

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_user_code_retries
    ):
        """
        Run after the task context has been finalized.

        is_task_ok is set to False if the user code raised an exception that
        was not handled by any decorator.

        Note that you can't create or modify data artifacts in this method
        since the task has been finalized by the time this method
        is called. Also note that the task may fail after this method has been
        called, so this method may get called multiple times for a task over
        multiple attempts, similar to all task_ methods.
        """
        pass


def _base_flow_decorator(decofunc, *args, **kwargs):
    """
    Decorator prototype for all flow (class) decorators. This function gets
    specialized and imported for all decorators types by
    _import_plugin_decorators().
    """
    if args:
        # No keyword arguments specified for the decorator, e.g. @foobar.
        # The first argument is the class to be decorated.
        cls = args[0]
        if isinstance(cls, type) and issubclass(cls, FlowSpec):
            # flow decorators add attributes in the class dictionary,
            # _flow_decorators. _flow_decorators is of type `{key:[decos]}`
            if decofunc.name in cls._flow_decorators and not decofunc.allow_multiple:
                raise DuplicateFlowDecoratorException(decofunc.name)
            else:
                deco_instance = decofunc(attributes=kwargs, statically_defined=True)
                cls._flow_decorators.setdefault(decofunc.name, []).append(deco_instance)
        else:
            raise BadFlowDecoratorException(decofunc.name)
        return cls
    else:
        # Keyword arguments specified, e.g. @foobar(a=1, b=2).
        # Return a decorator function that will get the actual
        # function to be decorated as the first argument.
        def wrap(f):
            return _base_flow_decorator(decofunc, f, **kwargs)

        return wrap


def _base_step_decorator(decotype, *args, **kwargs):
    """
    Decorator prototype for all step decorators. This function gets specialized
    and imported for all decorators types by _import_plugin_decorators().
    """

    if args:
        # No keyword arguments specified for the decorator, e.g. @foobar.
        # The first argument is the function to be decorated.
        func = args[0]
        if isinstance(func, (StepMutator, UserStepDecoratorBase)):
            func = func._my_step
        if not hasattr(func, "is_step"):
            raise BadStepDecoratorException(decotype.name, func)

        # if `allow_multiple` is not `True` then only one decorator type is allowed per step
        if (
            decotype.name in [deco.name for deco in func.decorators]
            and not decotype.allow_multiple
        ):
            raise DuplicateStepDecoratorException(decotype.name, func)
        else:
            func.decorators.append(decotype(attributes=kwargs, statically_defined=True))

        return func
    else:
        # Keyword arguments specified, e.g. @foobar(a=1, b=2).
        # Return a decorator function that will get the actual
        # function to be decorated as the first argument.
        def wrap(f):
            return _base_step_decorator(decotype, f, **kwargs)

        return wrap


_all_step_decos = None
_all_flow_decos = None


def get_all_step_decos():
    global _all_step_decos
    if _all_step_decos is None:
        from .plugins import STEP_DECORATORS

        _all_step_decos = {decotype.name: decotype for decotype in STEP_DECORATORS}
    return _all_step_decos


def get_all_flow_decos():
    global _all_flow_decos
    if _all_flow_decos is None:
        from .plugins import FLOW_DECORATORS

        _all_flow_decos = {decotype.name: decotype for decotype in FLOW_DECORATORS}
    return _all_flow_decos


def extract_step_decorator_from_decospec(decospec: str):
    splits = decospec.split(":", 1)
    deconame = splits[0]

    # Check if it is a user-defined decorator or metaflow decorator
    deco_cls = UserStepDecoratorMeta.get_decorator_by_name(deconame)
    if deco_cls is not None:
        return (
            deco_cls.parse_decorator_spec(splits[1] if len(splits) > 1 else ""),
            len(splits) > 1,
        )

    # Check if this is a decorator we can import
    if "." in deconame:
        # We consider this to be a import path to a user decorator so
        # something like "my_package.my_decorator"
        module_name, class_name = deconame.rsplit(".", 1)
        try:
            module = importlib.import_module(module_name)
        except ImportError as e:
            raise MetaflowException(
                "Could not import user decorator %s" % deconame
            ) from e
        deco_cls = getattr(module, class_name, None)
        if (
            deco_cls is None
            or not isinstance(deco_cls, type)
            or not issubclass(deco_cls, UserStepDecoratorBase)
        ):
            raise UnknownStepDecoratorException(deconame)
        return (
            deco_cls.parse_decorator_spec(splits[1] if len(splits) > 1 else ""),
            len(splits) > 1,
        )

    raise UnknownStepDecoratorException(deconame)


def extract_flow_decorator_from_decospec(decospec: str):
    splits = decospec.split(":", 1)
    deconame = splits[0]
    # Check if it is a user-defined decorator or metaflow decorator
    deco_cls = FlowMutatorMeta.get_decorator_by_name(deconame)
    if deco_cls is not None:
        return (
            deco_cls.parse_decorator_spec(splits[1] if len(splits) > 1 else ""),
            len(splits) > 1,
        )
    else:
        raise UnknownFlowDecoratorException(deconame)


def _attach_decorators(flow, decospecs):
    """
    Attach decorators to all steps during runtime. This has the same
    effect as if you defined the decorators statically in the source for
    every step. Used by --with command line parameter.
    """
    # Attach the decorator to all steps that don't have this decorator
    # already. This means that statically defined decorators are always
    # preferred over runtime decorators.
    #
    # Note that each step gets its own instance of the decorator class,
    # so decorator can maintain step-specific state.

    for step in flow:
        _attach_decorators_to_step(step, decospecs)


def _attach_decorators_to_step(step, decospecs):
    """
    Attach decorators to a step during runtime. This has the same
    effect as if you defined the decorators statically in the source for
    the step.
    """
    for decospec in decospecs:
        step_deco, _ = extract_step_decorator_from_decospec(decospec)
        if isinstance(step_deco, StepDecorator):
            # Check multiple
            if (
                step_deco.name not in [deco.name for deco in step.decorators]
                or step_deco.allow_multiple
            ):
                step.decorators.append(step_deco)
            # Else it is ignored -- this is a non-static decorator

        else:
            step_deco.add_or_raise(step, False, 1, None)


def _should_skip_decorator_for_spin(
    deco, is_spin, skip_decorators, logger, decorator_type="decorator"
):
    """
    Determine if a decorator should be skipped for spin steps.

    Parameters:
    -----------
    deco : Decorator
        The decorator instance to check
    is_spin : bool
        Whether this is a spin step
    skip_decorators : bool
        Whether to skip all decorators
    logger : callable
        Logger function for warnings
    decorator_type : str
        Type of decorator ("Flow decorator" or "Step decorator") for logging

    Returns:
    --------
    bool
        True if the decorator should be skipped, False otherwise
    """
    if not is_spin:
        return False

    # Skip all decorator hooks if skip_decorators is True
    if skip_decorators:
        return True

    # Run decorator hooks for spin steps only if they are in the whitelist
    if deco.name not in SPIN_ALLOWED_DECORATORS:
        logger(
            f"[Warning] Ignoring {decorator_type} '{deco.name}' as it is not supported in spin steps.",
            system_msg=True,
            timestamp=False,
            bad=True,
        )
        return True

    return False


def _init(flow, only_non_static=False):
    for decorators in flow._flow_decorators.values():
        for deco in decorators:
            deco.external_init()

    for flowstep in flow:
        for deco in flowstep.decorators:
            deco.external_init()
        for deco in flowstep.config_decorators or []:
            deco.external_init()
        for deco in flowstep.wrappers or []:
            deco.external_init()


def _init_flow_decorators(
    flow,
    graph,
    environment,
    flow_datastore,
    metadata,
    logger,
    echo,
    deco_options,
    is_spin=False,
    skip_decorators=False,
):
    # Since all flow decorators are stored as `{key:[deco]}` we iterate through each of them.
    for decorators in flow._flow_decorators.values():
        # First resolve the `options` for the flow decorator.
        # Options are passed from cli.
        # For example `@project` can take a `--name` / `--branch` from the cli as options.
        deco_flow_init_options = {}
        deco = decorators[0]
        # If a flow decorator allow multiple of same type then we don't allow multiple options for it.
        if deco.allow_multiple:
            if len(deco.options) > 0:
                raise MetaflowException(
                    "Flow decorator `@%s` has multiple options, which is not allowed. "
                    "Please ensure the FlowDecorator `%s` has no options since flow decorators with "
                    "`allow_mutiple=True` are not allowed to have options"
                    % (deco.name, deco.__class__.__name__)
                )
        else:
            # Each "non-multiple" flow decorator is only allowed to have one set of options
            # Note that there may be no deco_options if a MutableFlow config injected
            # the decorator.
            deco_flow_init_options = {
                option: deco_options.get(
                    option.replace("-", "_"), option_info["default"]
                )
                for option, option_info in deco.options.items()
            }
        for deco in decorators:
            if _should_skip_decorator_for_spin(
                deco, is_spin, skip_decorators, logger, "Flow decorator"
            ):
                continue
            deco.flow_init(
                flow,
                graph,
                environment,
                flow_datastore,
                metadata,
                logger,
                echo,
                deco_flow_init_options,
            )


def _init_step_decorators(
    flow,
    graph,
    environment,
    flow_datastore,
    logger,
    is_spin=False,
    skip_decorators=False,
):
    # NOTE: We don't need the graph but keeping it for backwards compatibility with
    # extensions that use it directly. We will remove it at some point.

    # We call the mutate method for both the flow and step mutators.
    cls = flow.__class__
    # Run all the decorators. We first run the flow-level decorators
    # and then the step level ones to maintain a consistent order with how
    # other decorators are run.

    for deco in cls._flow_state.get(_FlowState.FLOW_MUTATORS, []):
        if isinstance(deco, FlowMutator):
            inserted_by_value = [deco.decorator_name] + (deco.inserted_by or [])
            mutable_flow = MutableFlow(
                cls,
                pre_mutate=False,
                statically_defined=deco.statically_defined,
                inserted_by=inserted_by_value,
            )
            # Sanity check to make sure we are applying the decorator to the right
            # class
            if not deco._flow_cls == cls and not issubclass(cls, deco._flow_cls):
                raise MetaflowInternalError(
                    "FlowMutator registered on the wrong flow -- "
                    "expected %s but got %s" % (deco._flow_cls.__name__, cls.__name__)
                )
            debug.userconf_exec(
                "Evaluating flow level decorator %s (mutate)" % deco.__class__.__name__
            )
            deco.mutate(mutable_flow)
            # We reset cached_parameters on the very off chance that the user added
            # more configurations based on the configuration
            if _FlowState.CACHED_PARAMETERS in cls._flow_state:
                del cls._flow_state[_FlowState.CACHED_PARAMETERS]
        else:
            raise MetaflowInternalError(
                "A non FlowMutator found in flow custom decorators"
            )

    for step in cls._steps:
        for deco in step.config_decorators:
            inserted_by_value = [deco.decorator_name] + (deco.inserted_by or [])

            if isinstance(deco, StepMutator):
                debug.userconf_exec(
                    "Evaluating step level decorator %s for %s (mutate)"
                    % (deco.__class__.__name__, step.name)
                )
                deco.mutate(
                    MutableStep(
                        cls,
                        step,
                        pre_mutate=False,
                        statically_defined=deco.statically_defined,
                        inserted_by=inserted_by_value,
                    )
                )
            else:
                raise MetaflowInternalError(
                    "A non StepMutator found in step custom decorators"
                )

        if step.config_decorators:
            # We remove all mention of the custom step decorator
            setattr(cls, step.name, step)

    cls._init_graph()
    graph = flow._graph

    for step in flow:
        for deco in step.decorators:
            if _should_skip_decorator_for_spin(
                deco, is_spin, skip_decorators, logger, "Step decorator"
            ):
                continue
            deco.step_init(
                flow,
                graph,
                step.__name__,
                step.decorators,
                environment,
                flow_datastore,
                logger,
            )


FlowSpecDerived = TypeVar("FlowSpecDerived", bound=FlowSpec)

# The StepFlag is a "fake" input item to be able to distinguish
# callables and those that have had a `@step` decorator on them. This enables us
# to check the ordering of decorators (ie: put @step first) with the type
# system. There should be a better way to do this with a more flexible type
# system but this is what works for now with the Python type system
StepFlag = NewType("StepFlag", bool)


@overload
def step(
    f: Callable[[FlowSpecDerived], None],
) -> Callable[[FlowSpecDerived, StepFlag], None]: ...


@overload
def step(
    f: Callable[[FlowSpecDerived, Any], None],
) -> Callable[[FlowSpecDerived, Any, StepFlag], None]: ...


def step(
    f: Union[Callable[[FlowSpecDerived], None], Callable[[FlowSpecDerived, Any], None]],
):
    """
    Marks a method in a FlowSpec as a Metaflow Step. Note that this
    decorator needs to be placed as close to the method as possible (ie:
    before other decorators).

    In other words, this is valid:
    ```
    @batch
    @step
    def foo(self):
        pass
    ```

    whereas this is not:
    ```
    @step
    @batch
    def foo(self):
        pass
    ```

    Parameters
    ----------
    f : Union[Callable[[FlowSpecDerived], None], Callable[[FlowSpecDerived, Any], None]]
        Function to make into a Metaflow Step

    Returns
    -------
    Union[Callable[[FlowSpecDerived, StepFlag], None], Callable[[FlowSpecDerived, Any, StepFlag], None]]
        Function that is a Metaflow Step
    """
    f.is_step = True
    f.decorators = []
    f.config_decorators = []
    f.wrappers = []
    f.name = f.__name__
    return f


def _import_plugin_decorators(globals_dict):
    """
    Auto-generate a decorator function for every decorator
    defined in plugins.STEP_DECORATORS and plugins.FLOW_DECORATORS.
    """
    from .plugins import STEP_DECORATORS, FLOW_DECORATORS

    # Q: Why not use StepDecorators directly as decorators?
    # A: Getting an object behave as a decorator that can work
    #    both with and without arguments is surprisingly hard.
    #    It is easier to make plain function decorators work in
    #    the dual mode - see _base_step_decorator above.
    for decotype in STEP_DECORATORS:
        globals_dict[decotype.name] = partial(_base_step_decorator, decotype)

    # add flow-level decorators
    for decotype in FLOW_DECORATORS:
        globals_dict[decotype.name] = partial(_base_flow_decorator, decotype)



import json

from functools import wraps

from metaflow._vendor import click

from .. import decorators, namespace, parameters, tracing
from ..exception import CommandException
from ..graph import FlowGraph
from ..metaflow_current import current
from ..metaflow_config import (
    DEFAULT_DECOSPECS,
    FEAT_ALWAYS_UPLOAD_CODE_PACKAGE,
    SPIN_PERSIST,
)
from ..metaflow_profile import from_start
from ..package import MetaflowPackage
from ..runtime import NativeRuntime, SpinRuntime
from ..system import _system_logger

# from ..client.core import Run

from ..tagging_util import validate_tags
from ..util import get_latest_run_id, write_latest_run_id, parse_spin_pathspec


def before_run(obj, tags, decospecs, skip_decorators=False):
    validate_tags(tags)

    # There's a --with option both at the top-level and for the run/resume/spin
    # subcommand. Why?
    #
    # "run --with shoes" looks so much better than "--with shoes run".
    # This is a very common use case of --with.
    #
    # A downside is that we need to have the following decorators handling
    # in two places in this module and make sure _init_step_decorators
    # doesn't get called twice.

    # We want the order to be the following:
    # - run level decospecs
    # - top level decospecs
    # - environment decospecs
    from_start(
        f"Inside before_run, skip_decorators={skip_decorators}, is_spin={obj.is_spin}"
    )
    if not skip_decorators:
        all_decospecs = (
            list(decospecs or [])
            + obj.tl_decospecs
            + list(obj.environment.decospecs() or [])
        )
        if all_decospecs:
            # These decospecs are the ones from run/resume/spin PLUS the ones from the
            # environment (for example the @conda)
            decorators._attach_decorators(obj.flow, all_decospecs)
            decorators._init(obj.flow)
            # Regenerate graph if we attached more decorators
            obj.flow.__class__._init_graph()
            obj.graph = obj.flow._graph

        obj.check(obj.graph, obj.flow, obj.environment, pylint=obj.pylint)
        # obj.environment.init_environment(obj.logger)

        decorators._init_step_decorators(
            obj.flow,
            obj.graph,
            obj.environment,
            obj.flow_datastore,
            obj.logger,
            obj.is_spin,
            skip_decorators,
        )
    # Re-read graph since it may have been modified by mutators
    obj.graph = obj.flow._graph

    obj.metadata.add_sticky_tags(tags=tags)

    # Package working directory only once per run.
    # We explicitly avoid doing this in `start` since it is invoked for every
    # step in the run.
    obj.package = MetaflowPackage(
        obj.flow,
        obj.environment,
        obj.echo,
        suffixes=obj.package_suffixes,
        flow_datastore=obj.flow_datastore if FEAT_ALWAYS_UPLOAD_CODE_PACKAGE else None,
    )


def common_runner_options(func):
    @click.option(
        "--run-id-file",
        default=None,
        show_default=True,
        type=str,
        help="Write the ID of this run to the file specified.",
    )
    @click.option(
        "--runner-attribute-file",
        default=None,
        show_default=True,
        type=str,
        help="Write the metadata and pathspec of this run to the file specified. Used internally "
        "for Metaflow's Runner API.",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def write_file(file_path, content):
    if file_path is not None:
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(str(content))


def config_callback(ctx, param, value):
    # Callback to:
    #  - read  the Click auto_envvar variable from both the
    #    environment AND the configuration
    #  - merge that value with the value passed in the command line (value)
    #  - return the value as a tuple
    # Note that this function gets called even if there is no option passed on the
    # command line.
    # NOTE: Assumes that ctx.auto_envvar_prefix is set to METAFLOW (same as in
    # from_conf)

    # Read decospecs options from the environment (METAFLOW_DEFAULT_DECOSPECS=...)
    # and merge them with the one provided as --with.
    splits = DEFAULT_DECOSPECS.split()
    return tuple(list(value) + splits)


def common_run_options(func):
    @click.option(
        "--tag",
        "tags",
        multiple=True,
        default=None,
        help="Annotate this run with the given tag. You can specify "
        "this option multiple times to attach multiple tags in "
        "the run.",
    )
    @click.option(
        "--max-workers",
        default=16,
        show_default=True,
        help="Maximum number of parallel processes.",
    )
    @click.option(
        "--max-num-splits",
        default=100,
        show_default=True,
        help="Maximum number of splits allowed in a foreach. This "
        "is a safety check preventing bugs from triggering "
        "thousands of steps inadvertently.",
    )
    @click.option(
        "--max-log-size",
        default=10,
        show_default=True,
        help="Maximum size of stdout and stderr captured in "
        "megabytes. If a step outputs more than this to "
        "stdout/stderr, its output will be truncated.",
    )
    @click.option(
        "--with",
        "decospecs",
        multiple=True,
        help="Add a decorator to all steps. You can specify this "
        "option multiple times to attach multiple decorators "
        "in steps.",
        callback=config_callback,
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.option(
    "--origin-run-id",
    default=None,
    help="ID of the run that should be resumed. By default, the "
    "last run executed locally.",
)
@click.option(
    "--run-id",
    default=None,
    help="Run ID for the new run. By default, a new run-id will be generated",
    hidden=True,
)
@click.option(
    "--clone-only/--no-clone-only",
    default=False,
    show_default=True,
    help="Only clone tasks without continuing execution",
    hidden=True,
)
@click.option(
    "--reentrant/--no-reentrant",
    default=False,
    show_default=True,
    hidden=True,
    help="If specified, allows this call to be called in parallel",
)
@click.option(
    "--resume-identifier",
    default=None,
    show_default=True,
    hidden=True,
    help="If specified, it identifies the task that started this resume call. It is in the form of {step_name}-{task_id}",
)
@click.argument("step-to-rerun", required=False)
@click.command(help="Resume execution of a previous run of this flow.")
@tracing.cli("cli/resume")
@common_run_options
@common_runner_options
@click.pass_obj
def resume(
    obj,
    tags=None,
    step_to_rerun=None,
    origin_run_id=None,
    run_id=None,
    clone_only=False,
    reentrant=False,
    max_workers=None,
    max_num_splits=None,
    max_log_size=None,
    decospecs=None,
    run_id_file=None,
    resume_identifier=None,
    runner_attribute_file=None,
):
    before_run(obj, tags, decospecs)

    if origin_run_id is None:
        origin_run_id = get_latest_run_id(obj.echo, obj.flow.name)
        if origin_run_id is None:
            raise CommandException(
                "A previous run id was not found. Specify --origin-run-id."
            )

    if step_to_rerun is None:
        steps_to_rerun = set()
    else:
        # validate step name
        if step_to_rerun not in obj.graph.nodes:
            raise CommandException(
                "invalid step name {0} specified, must be step present in "
                "current form of execution graph. Valid step names include: {1}".format(
                    step_to_rerun, ",".join(list(obj.graph.nodes.keys()))
                )
            )

        ## TODO: instead of checking execution path here, can add a warning later
        ## instead of throwing an error. This is for resuming a step which was not
        ## taken inside a branch i.e. not present in the execution path.

        # origin_run = Run(f"{obj.flow.name}/{origin_run_id}", _namespace_check=False)
        # executed_steps = {step.path_components[-1] for step in origin_run}
        # if step_to_rerun not in executed_steps:
        #     raise CommandException(
        #         f"Cannot resume from step '{step_to_rerun}'. This step was not "
        #         f"part of the original execution path for run '{origin_run_id}'."
        #     )

        steps_to_rerun = {step_to_rerun}

    if run_id:
        # Run-ids that are provided by the metadata service are always integers.
        # External providers or run-ids (like external schedulers) always need to
        # be non-integers to avoid any clashes. This condition ensures this.
        try:
            int(run_id)
        except:
            pass
        else:
            raise CommandException("run-id %s cannot be an integer" % run_id)

    runtime = NativeRuntime(
        obj.flow,
        obj.graph,
        obj.flow_datastore,
        obj.metadata,
        obj.environment,
        obj.package,
        obj.logger,
        obj.entrypoint,
        obj.event_logger,
        obj.monitor,
        run_id=run_id,
        clone_run_id=origin_run_id,
        clone_only=clone_only,
        reentrant=reentrant,
        steps_to_rerun=steps_to_rerun,
        max_workers=max_workers,
        max_num_splits=max_num_splits,
        max_log_size=max_log_size * 1024 * 1024,
        resume_identifier=resume_identifier,
    )
    write_file(run_id_file, runtime.run_id)
    runtime.print_workflow_info()

    runtime.persist_constants()

    if runner_attribute_file:
        with open(runner_attribute_file, "w", encoding="utf-8") as f:
            json.dump(
                {
                    "run_id": runtime.run_id,
                    "flow_name": obj.flow.name,
                    "metadata": obj.metadata.metadata_str(),
                },
                f,
            )

    # We may skip clone-only resume if this is not a resume leader,
    # and clone is already complete.
    if runtime.should_skip_clone_only_execution():
        return

    current._update_env(
        {
            "run_id": runtime.run_id,
        }
    )
    _system_logger.log_event(
        level="info",
        module="metaflow.resume",
        name="start",
        payload={
            "msg": "Resuming run",
        },
    )

    with runtime.run_heartbeat():
        if clone_only:
            runtime.clone_original_run()
        else:
            runtime.clone_original_run(generate_task_obj=True, verbose=False)
            runtime.execute()


@parameters.add_custom_parameters(deploy_mode=True)
@click.command(help="Run the workflow locally.")
@tracing.cli("cli/run")
@common_run_options
@common_runner_options
@click.option(
    "--namespace",
    "user_namespace",
    default=None,
    help="Change namespace from the default (your username) to "
    "the specified tag. Note that this option does not alter "
    "tags assigned to the objects produced by this run, just "
    "what existing objects are visible in the client API. You "
    "can enable the global namespace with an empty string."
    "--namespace=",
)
@click.pass_obj
def run(
    obj,
    tags=None,
    max_workers=None,
    max_num_splits=None,
    max_log_size=None,
    decospecs=None,
    run_id_file=None,
    runner_attribute_file=None,
    user_namespace=None,
    **kwargs,
):
    if user_namespace is not None:
        namespace(user_namespace or None)
    before_run(obj, tags, decospecs)

    runtime = NativeRuntime(
        obj.flow,
        obj.graph,
        obj.flow_datastore,
        obj.metadata,
        obj.environment,
        obj.package,
        obj.logger,
        obj.entrypoint,
        obj.event_logger,
        obj.monitor,
        max_workers=max_workers,
        max_num_splits=max_num_splits,
        max_log_size=max_log_size * 1024 * 1024,
    )
    write_latest_run_id(obj, runtime.run_id)
    write_file(run_id_file, runtime.run_id)

    obj.flow._set_constants(obj.graph, kwargs, obj.config_options)
    current._update_env(
        {
            "run_id": runtime.run_id,
        }
    )
    _system_logger.log_event(
        level="info",
        module="metaflow.run",
        name="start",
        payload={
            "msg": "Starting run",
        },
    )

    runtime.print_workflow_info()
    runtime.persist_constants()
    if runner_attribute_file:
        with open(runner_attribute_file, "w", encoding="utf-8") as f:
            json.dump(
                {
                    "run_id": runtime.run_id,
                    "flow_name": obj.flow.name,
                    "metadata": obj.metadata.metadata_str(),
                },
                f,
            )
    with runtime.run_heartbeat():
        runtime.execute()


# @parameters.add_custom_parameters(deploy_mode=True)
@click.command(help="Spins up a task for a given step from a previous run locally.")
@tracing.cli("cli/spin")
@click.argument("pathspec")
@click.option(
    "--skip-decorators/--no-skip-decorators",
    is_flag=True,
    # Default False matches the saved_args check in cli.py for spin steps - skip_decorators
    # only becomes True when explicitly passed, otherwise decorators are applied by default
    default=False,
    show_default=True,
    help="Skip decorators attached to the step or flow.",
)
@click.option(
    "--artifacts-module",
    default=None,
    show_default=True,
    help="Path to a module that contains artifacts to be used in the spun step. "
    "The artifacts should be defined as a dictionary called ARTIFACTS with keys as "
    "the artifact names and values as the artifact values. The artifact values will "
    "overwrite the default values of the artifacts used in the spun step.",
)
@click.option(
    "--persist/--no-persist",
    "persist",
    default=SPIN_PERSIST,
    show_default=True,
    help="Whether to persist the artifacts in the spun step. If set to False, "
    "the artifacts will not be persisted and will not be available in the spun step's "
    "datastore.",
)
@click.option(
    "--max-log-size",
    default=10,
    show_default=True,
    help="Maximum size of stdout and stderr captured in "
    "megabytes. If a step outputs more than this to "
    "stdout/stderr, its output will be truncated.",
)
@common_runner_options
@click.pass_obj
def spin(
    obj,
    pathspec,
    persist=True,
    artifacts_module=None,
    skip_decorators=False,
    max_log_size=None,
    run_id_file=None,
    runner_attribute_file=None,
    **kwargs,
):
    # Parse the pathspec argument to extract step name and full pathspec
    step_name, parsed_pathspec = parse_spin_pathspec(pathspec, obj.flow.name)

    before_run(obj, [], [], skip_decorators)
    obj.echo(f"Spinning up step *{step_name}* locally for flow *{obj.flow.name}*")
    # For spin, flow parameters come from the original run, but _set_constants
    # requires them in kwargs. Use parameter defaults as placeholders - they'll be
    # overwritten when the spin step loads artifacts from the original run.
    flow_param_defaults = {}
    for var, param in obj.flow._get_parameters():
        if not param.IS_CONFIG_PARAMETER:
            default_value = param.kwargs.get("default")
            # Use None for required parameters without defaults
            flow_param_defaults[param.name.replace("-", "_").lower()] = default_value
    obj.flow._set_constants(obj.graph, flow_param_defaults, obj.config_options)
    step_func = getattr(obj.flow, step_name, None)
    if step_func is None:
        raise CommandException(
            f"Step '{step_name}' not found in flow '{obj.flow.name}'. "
            "Please provide a valid step name."
        )
    from_start("Spin: before spin runtime init")
    spin_runtime = SpinRuntime(
        obj.flow,
        obj.graph,
        obj.flow_datastore,
        obj.metadata,
        obj.environment,
        obj.package,
        obj.logger,
        obj.entrypoint,
        obj.event_logger,
        obj.monitor,
        step_func,
        step_name,
        parsed_pathspec,
        skip_decorators,
        artifacts_module,
        persist,
        max_log_size * 1024 * 1024,
    )
    write_latest_run_id(obj, spin_runtime.run_id)
    write_file(run_id_file, spin_runtime.run_id)
    # We only need the root for the metadata, i.e. the portion before DATASTORE_LOCAL_DIR
    datastore_root = spin_runtime._flow_datastore._storage_impl.datastore_root
    orig_task_metadata_root = datastore_root.rsplit("/", 1)[0]
    from_start("Spin: going to execute")
    spin_runtime.execute()
    from_start("Spin: after spin runtime execute")

    if runner_attribute_file:
        with open(runner_attribute_file, "w") as f:
            json.dump(
                {
                    "task_id": spin_runtime.task.task_id,
                    "step_name": step_name,
                    "run_id": spin_runtime.run_id,
                    "flow_name": obj.flow.name,
                    # Store metadata in a format that can be used by the Runner API
                    "metadata": f"{obj.metadata.__class__.TYPE}@{orig_task_metadata_root}",
                },
                f,
            )



import json
import os
import sys

from os.path import expanduser

from metaflow.util import to_unicode
from metaflow._vendor import click
from metaflow.util import to_unicode


from .util import echo_always, makedirs


echo = echo_always

# NOTE: This code needs to be in sync with metaflow/metaflow_config.py.
METAFLOW_CONFIGURATION_DIR = expanduser(
    os.environ.get("METAFLOW_HOME", "~/.metaflowconfig")
)
METAFLOW_PROFILE = os.environ.get("METAFLOW_PROFILE", "")


@click.group()
def cli():
    pass


@cli.group(help="Configure Metaflow to access the cloud.")
def configure():
    makedirs(METAFLOW_CONFIGURATION_DIR)


def get_config_path(profile):
    config_file = "config.json" if not profile else ("config_%s.json" % profile)
    path = os.path.join(METAFLOW_CONFIGURATION_DIR, config_file)
    return path


def confirm_overwrite_config(profile):
    path = get_config_path(profile)
    if os.path.exists(path):
        if not click.confirm(
            click.style(
                "We found an existing configuration for your "
                + "profile. Do you want to modify the existing "
                + "configuration?",
                fg="red",
                bold=True,
            )
        ):
            echo(
                "You can configure a different named profile by using the "
                "--profile argument. You can activate this profile by setting "
                "the environment variable METAFLOW_PROFILE to the named "
                "profile.",
                fg="yellow",
            )
            return False
    return True


def check_for_missing_profile(profile):
    path = get_config_path(profile)
    # Absence of default config is equivalent to running locally.
    if profile and not os.path.exists(path):
        raise click.ClickException(
            "Couldn't find configuration for profile "
            + click.style('"%s"' % profile, fg="red")
            + " in "
            + click.style('"%s"' % path, fg="red")
        )


def get_env(profile):
    path = get_config_path(profile)
    if os.path.exists(path):
        with open(path) as f:
            return json.load(f)
    return {}


def persist_env(env_dict, profile):
    # TODO: Should we persist empty env_dict or notify user differently?
    path = get_config_path(profile)

    with open(path, "w") as f:
        json.dump(env_dict, f, indent=4, sort_keys=True)

    echo("\nConfiguration successfully written to ", nl=False, bold=True)
    echo('"%s"' % path, fg="cyan")


@configure.command(help="Reset configuration to disable cloud access.")
@click.option(
    "--profile", "-p", default=METAFLOW_PROFILE, help="Optional named profile."
)
def reset(profile):
    check_for_missing_profile(profile)
    path = get_config_path(profile)
    if os.path.exists(path):
        if click.confirm(
            "Do you really wish to reset the configuration in "
            + click.style('"%s"' % path, fg="cyan"),
            abort=True,
        ):
            os.remove(path)
            echo("Configuration successfully reset to run locally.")
    else:
        echo("Configuration is already reset to run locally.")


@configure.command(help="Show existing configuration.")
@click.option(
    "--profile", "-p", default=METAFLOW_PROFILE, help="Optional named profile."
)
def show(profile):
    check_for_missing_profile(profile)
    path = get_config_path(profile)
    env_dict = {}
    if os.path.exists(path):
        with open(path, "r") as f:
            env_dict = json.load(f)
    if env_dict:
        echo("Showing configuration in ", nl=False)
        echo('"%s"\n' % path, fg="cyan")
        for k, v in env_dict.items():
            echo("%s=%s" % (k, v))
    else:
        echo("Configuration is set to run locally.")


@configure.command(help="Export configuration to a file.")
@click.option(
    "--profile",
    "-p",
    default=METAFLOW_PROFILE,
    help="Optional named profile whose configuration must be " "exported.",
)
@click.argument("output_filename", type=click.Path(resolve_path=True))
def export(profile, output_filename):
    check_for_missing_profile(profile)
    # Export its contents to a new file.
    path = get_config_path(profile)
    env_dict = {}
    if os.path.exists(path):
        with open(path, "r") as f:
            env_dict = json.load(f)
    # resolve_path doesn't expand `~` in `path`.
    output_path = expanduser(output_filename)
    if os.path.exists(output_path):
        if click.confirm(
            "Do you wish to overwrite the contents in "
            + click.style('"%s"' % output_path, fg="cyan")
            + "?",
            abort=True,
        ):
            pass
    # Write to file.
    with open(output_path, "w") as f:
        json.dump(env_dict, f, indent=4, sort_keys=True)
    echo("Configuration successfully exported to: ", nl=False)
    echo('"%s"' % output_path, fg="cyan")


@configure.command(help="Import configuration from a file.", name="import")
@click.option(
    "--profile",
    "-p",
    default=METAFLOW_PROFILE,
    help="Optional named profile to which the configuration must be " "imported into.",
)
@click.argument("input_filename", type=click.Path(exists=True, resolve_path=True))
def import_from(profile, input_filename):
    check_for_missing_profile(profile)
    # Import configuration.
    input_path = expanduser(input_filename)
    env_dict = {}
    with open(input_path, "r") as f:
        env_dict = json.load(f)
    echo("Configuration successfully read from: ", nl=False)
    echo('"%s"' % input_path, fg="cyan")

    # Persist configuration.
    confirm_overwrite_config(profile)
    persist_env(env_dict, profile)


@configure.command(help="Configure metaflow to access hosted sandbox.")
@click.option(
    "--profile",
    "-p",
    default="",
    help="Configure a named profile. Activate the profile by setting "
    "`METAFLOW_PROFILE` environment variable.",
)
@click.option(
    "--overwrite/--no-overwrite",
    "-o/",
    default=False,
    show_default=True,
    help="Overwrite profile configuration without asking",
)
def sandbox(profile, overwrite):
    if not overwrite:
        confirm_overwrite_config(profile)
    # Prompt for user input.
    encoded_str = click.prompt(
        "Following instructions from "
        "https://metaflow.org/sandbox, "
        "please paste the encoded magic string",
        type=str,
    )
    # Decode the bytes to env_dict.
    try:
        import base64
        import zlib
        from metaflow.util import to_bytes

        env_dict = json.loads(
            to_unicode(zlib.decompress(base64.b64decode(to_bytes(encoded_str))))
        )
    except:
        # TODO: Add the URL for contact us page in the error?
        raise click.BadArgumentUsage(
            "Could not decode the sandbox " "configuration. Please contact us."
        )
    # Persist to a file.
    persist_env(env_dict, profile)


def cyan(string):
    return click.style(string, fg="cyan")


def yellow(string):
    return click.style(string, fg="yellow")


def red(string):
    return click.style(string, fg="red")


def configure_s3_datastore(existing_env):
    env = {}
    # Set Amazon S3 as default datastore.
    env["METAFLOW_DEFAULT_DATASTORE"] = "s3"
    # Set Amazon S3 folder for datastore.
    env["METAFLOW_DATASTORE_SYSROOT_S3"] = click.prompt(
        cyan("[METAFLOW_DATASTORE_SYSROOT_S3]")
        + " Amazon S3 folder for Metaflow artifact storage "
        + "(s3://<bucket>/<prefix>).",
        default=existing_env.get("METAFLOW_DATASTORE_SYSROOT_S3"),
        show_default=True,
    )
    # Set Amazon S3 folder for datatools.
    env["METAFLOW_DATATOOLS_S3ROOT"] = click.prompt(
        cyan("[METAFLOW_DATATOOLS_S3ROOT]")
        + yellow(" (optional)")
        + " Amazon S3 folder for Metaflow datatools "
        + "(s3://<bucket>/<prefix>).",
        default=existing_env.get(
            "METAFLOW_DATATOOLS_S3ROOT",
            os.path.join(env["METAFLOW_DATASTORE_SYSROOT_S3"], "data"),
        ),
        show_default=True,
    )
    return env


def configure_azure_datastore(existing_env):
    env = {}
    # Set Azure Blob Storage as default datastore.
    env["METAFLOW_DEFAULT_DATASTORE"] = "azure"
    # Set Azure Blob Storage folder for datastore.
    # TODO rename this Blob Endpoint!
    env["METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT"] = click.prompt(
        cyan("[METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT]")
        + " Azure Storage Account URL, for the account holding the Blob container to be used. "
        + "(E.g. https://<storage_account>.blob.core.windows.net/)",
        default=existing_env.get("METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT"),
        show_default=True,
    )
    env["METAFLOW_DATASTORE_SYSROOT_AZURE"] = click.prompt(
        cyan("[METAFLOW_DATASTORE_SYSROOT_AZURE]")
        + " Azure Blob Storage folder for Metaflow artifact storage "
        + "(Format: <container_name>/<prefix>)",
        default=existing_env.get("METAFLOW_DATASTORE_SYSROOT_AZURE"),
        show_default=True,
    )
    return env


def configure_gs_datastore(existing_env):
    env = {}
    # Set Google Cloud Storage as default datastore.
    env["METAFLOW_DEFAULT_DATASTORE"] = "gs"
    # Set Google Cloud Storage folder for datastore.
    env["METAFLOW_DATASTORE_SYSROOT_GS"] = click.prompt(
        cyan("[METAFLOW_DATASTORE_SYSROOT_GS]")
        + " Google Cloud Storage folder for Metaflow artifact storage "
        + "(Format: gs://<bucket>/<prefix>)",
        default=existing_env.get("METAFLOW_DATASTORE_SYSROOT_GS"),
        show_default=True,
    )
    return env


def configure_metadata_service(existing_env):
    empty_profile = False
    if not existing_env:
        empty_profile = True
    env = {}

    # Set Metadata Service as default.
    env["METAFLOW_DEFAULT_METADATA"] = "service"
    # Set URL for the Metadata Service.
    env["METAFLOW_SERVICE_URL"] = click.prompt(
        cyan("[METAFLOW_SERVICE_URL]") + " URL for Metaflow Service.",
        default=existing_env.get("METAFLOW_SERVICE_URL"),
        show_default=True,
    )
    # Set internal URL for the Metadata Service.
    env["METAFLOW_SERVICE_INTERNAL_URL"] = click.prompt(
        cyan("[METAFLOW_SERVICE_INTERNAL_URL]")
        + yellow(" (optional)")
        + " URL for Metaflow Service "
        + "(Accessible only within VPC [AWS] or a Kubernetes cluster [if the service runs in one]).",
        default=existing_env.get(
            "METAFLOW_SERVICE_INTERNAL_URL", env["METAFLOW_SERVICE_URL"]
        ),
        show_default=True,
    )
    # Set Auth Key for the Metadata Service.
    env["METAFLOW_SERVICE_AUTH_KEY"] = click.prompt(
        cyan("[METAFLOW_SERVICE_AUTH_KEY]")
        + yellow(" (optional)")
        + " Auth Key for Metaflow Service.",
        default=existing_env.get("METAFLOW_SERVICE_AUTH_KEY", ""),
        show_default=True,
    )
    return env


def configure_azure_datastore_and_metadata(existing_env):
    empty_profile = False
    if not existing_env:
        empty_profile = True
    env = {}

    # Configure Azure Blob Storage as the datastore.
    use_azure_as_datastore = click.confirm(
        "\nMetaflow can use "
        + yellow("Azure Blob Storage as the storage backend")
        + " for all code and data artifacts on "
        + "Azure.\nAzure Blob Storage is a strict requirement if you "
        + "intend to execute your flows on a Kubernetes cluster on Azure (AKS or self-managed)"
        + ".\nWould you like to configure Azure Blob Storage "
        + "as the default storage backend?",
        default=empty_profile
        or existing_env.get("METAFLOW_DEFAULT_DATASTORE", "") == "azure",
        abort=False,
    )
    if use_azure_as_datastore:
        env.update(configure_azure_datastore(existing_env))

    # Configure Metadata service for tracking.
    if click.confirm(
        "\nMetaflow can use a "
        + yellow("remote Metadata Service to track")
        + " and persist flow execution metadata.\nConfiguring the "
        "service is a requirement if you intend to schedule your "
        "flows with Kubernetes on Azure (AKS or self-managed).\nWould you like to "
        "configure the Metadata Service?",
        default=empty_profile
        or existing_env.get("METAFLOW_DEFAULT_METADATA", "") == "service",
        abort=False,
    ):
        env.update(configure_metadata_service(existing_env))
    return env


def configure_gs_datastore_and_metadata(existing_env):
    empty_profile = False
    if not existing_env:
        empty_profile = True
    env = {}

    # Configure Google Cloud Storage as the datastore.
    use_gs_as_datastore = click.confirm(
        "\nMetaflow can use "
        + yellow("Google Cloud Storage as the storage backend")
        + " for all code and data artifacts on "
        + "Google Cloud Storage.\nGoogle Cloud Storage is a strict requirement if you "
        + "intend to execute your flows on a Kubernetes cluster on GCP (GKE or self-managed)"
        + ".\nWould you like to configure Google Cloud Storage "
        + "as the default storage backend?",
        default=empty_profile
        or existing_env.get("METAFLOW_DEFAULT_DATASTORE", "") == "gs",
        abort=False,
    )
    if use_gs_as_datastore:
        env.update(configure_gs_datastore(existing_env))

    # Configure Metadata service for tracking.
    if click.confirm(
        "\nMetaflow can use a "
        + yellow("remote Metadata Service to track")
        + " and persist flow execution metadata.\nConfiguring the "
        "service is a requirement if you intend to schedule your "
        "flows with Kubernetes on GCP (GKE or self-managed).\nWould you like to "
        "configure the Metadata Service?",
        default=empty_profile
        or existing_env.get("METAFLOW_DEFAULT_METADATA", "") == "service",
        abort=False,
    ):
        env.update(configure_metadata_service(existing_env))
    return env


def configure_aws_datastore_and_metadata(existing_env):
    empty_profile = False
    if not existing_env:
        empty_profile = True
    env = {}

    # Configure Amazon S3 as the datastore.
    use_s3_as_datastore = click.confirm(
        "\nMetaflow can use "
        + yellow("Amazon S3 as the storage backend")
        + " for all code and data artifacts on "
        + "AWS.\nAmazon S3 is a strict requirement if you "
        + "intend to execute your flows on AWS Batch "
        + "and/or schedule them on AWS Step "
        + "Functions.\nWould you like to configure Amazon "
        + "S3 as the default storage backend?",
        default=empty_profile
        or existing_env.get("METAFLOW_DEFAULT_DATASTORE", "") == "s3",
        abort=False,
    )
    if use_s3_as_datastore:
        env.update(configure_s3_datastore(existing_env))

    # Configure Metadata service for tracking.
    if click.confirm(
        "\nMetaflow can use a "
        + yellow("remote Metadata Service to track")
        + " and persist flow execution metadata.\nConfiguring the "
        "service is a requirement if you intend to schedule your "
        "flows with AWS Step Functions.\nWould you like to "
        "configure the Metadata Service?",
        default=empty_profile
        or existing_env.get("METAFLOW_DEFAULT_METADATA", "") == "service"
        or "METAFLOW_SFN_IAM_ROLE" in env,
        abort=False,
    ):
        env.update(configure_metadata_service(existing_env))
    return env


def configure_aws_batch(existing_env):
    empty_profile = False
    if not existing_env:
        empty_profile = True
    env = {}

    # Set AWS Batch Job Queue.
    env["METAFLOW_BATCH_JOB_QUEUE"] = click.prompt(
        cyan("[METAFLOW_BATCH_JOB_QUEUE]") + " AWS Batch Job Queue.",
        default=existing_env.get("METAFLOW_BATCH_JOB_QUEUE"),
        show_default=True,
    )
    # Set IAM role for AWS Batch jobs to assume.
    env["METAFLOW_ECS_S3_ACCESS_IAM_ROLE"] = click.prompt(
        cyan("[METAFLOW_ECS_S3_ACCESS_IAM_ROLE]")
        + " IAM role for AWS Batch jobs to access AWS "
        + "resources (Amazon S3 etc.).",
        default=existing_env.get("METAFLOW_ECS_S3_ACCESS_IAM_ROLE"),
        show_default=True,
    )
    # Set default Docker repository for AWS Batch jobs.
    env["METAFLOW_BATCH_CONTAINER_REGISTRY"] = click.prompt(
        cyan("[METAFLOW_BATCH_CONTAINER_REGISTRY]")
        + yellow(" (optional)")
        + " Default Docker image repository for AWS "
        + "Batch jobs. If nothing is specified, "
        + "dockerhub (hub.docker.com/) is "
        + "used as default.",
        default=existing_env.get("METAFLOW_BATCH_CONTAINER_REGISTRY", ""),
        show_default=True,
    )
    # Set default Docker image for AWS Batch jobs.
    env["METAFLOW_BATCH_CONTAINER_IMAGE"] = click.prompt(
        cyan("[METAFLOW_BATCH_CONTAINER_IMAGE]")
        + yellow(" (optional)")
        + " Default Docker image for AWS Batch jobs. "
        + "If nothing is specified, an appropriate "
        + "python image is used as default.",
        default=existing_env.get("METAFLOW_BATCH_CONTAINER_IMAGE", ""),
        show_default=True,
    )

    # Configure AWS Step Functions for scheduling.
    if click.confirm(
        "\nMetaflow can "
        + yellow("schedule your flows on AWS Step " "Functions")
        + " and trigger them at a specific cadence using "
        "Amazon EventBridge.\nTo support flows involving "
        "foreach steps, you would need access to AWS "
        "DynamoDB.\nWould you like to configure AWS Step "
        "Functions for scheduling?",
        default=empty_profile or "METAFLOW_SFN_IAM_ROLE" in existing_env,
        abort=False,
    ):
        # Configure IAM role for AWS Step Functions.
        env["METAFLOW_SFN_IAM_ROLE"] = click.prompt(
            cyan("[METAFLOW_SFN_IAM_ROLE]")
            + " IAM role for AWS Step Functions to "
            + "access AWS resources (AWS Batch, "
            + "AWS DynamoDB).",
            default=existing_env.get("METAFLOW_SFN_IAM_ROLE"),
            show_default=True,
        )
        # Configure IAM role for AWS Events Bridge.
        env["METAFLOW_EVENTS_SFN_ACCESS_IAM_ROLE"] = click.prompt(
            cyan("[METAFLOW_EVENTS_SFN_ACCESS_IAM_ROLE]")
            + " IAM role for Amazon EventBridge to "
            + "access AWS Step Functions.",
            default=existing_env.get("METAFLOW_EVENTS_SFN_ACCESS_IAM_ROLE"),
            show_default=True,
        )
        # Configure AWS DynamoDB Table for AWS Step Functions.
        env["METAFLOW_SFN_DYNAMO_DB_TABLE"] = click.prompt(
            cyan("[METAFLOW_SFN_DYNAMO_DB_TABLE]")
            + " AWS DynamoDB table name for tracking "
            + "AWS Step Functions execution metadata.",
            default=existing_env.get("METAFLOW_SFN_DYNAMO_DB_TABLE"),
            show_default=True,
        )
    return env


def check_kubernetes_client(ctx):
    try:
        import kubernetes
    except ImportError:
        echo(
            "Could not import module 'Kubernetes'.\nInstall Kubernetes "
            + "Python package (https://pypi.org/project/kubernetes/) first.\n"
            "You can install the module by executing - \n"
            + yellow("%s -m pip install kubernetes" % sys.executable)
            + " \nor equivalent in your favorite Python package manager\n"
        )
        ctx.abort()


def check_kubernetes_config(ctx):
    from kubernetes import config

    try:
        all_contexts, current_context = config.list_kube_config_contexts()
        click.confirm(
            "You have a valid kubernetes configuration. The current context is set to "
            + yellow(current_context["name"])
            + " "
            + "Proceed?",
            default=True,
            abort=True,
        )
    except config.config_exception.ConfigException as e:
        click.confirm(
            "\nYou don't seem to have a valid Kubernetes configuration file. "
            + "The error from Kubernetes client library: "
            + red(str(e))
            + "."
            + "To create a kubernetes configuration for EKS, you typically need to run "
            + yellow("aws eks update-kubeconfig --name <CLUSTER NAME>")
            + ". For further details, refer to AWS documentation at https://docs.aws.amazon.com/eks/latest/userguide/create-kubeconfig.html\n"
            "Do you want to proceed with configuring Metaflow for Kubernetes anyway?",
            default=False,
            abort=True,
        )


def configure_argo_events(existing_env):
    env = {}

    # Argo events service account
    env["METAFLOW_ARGO_EVENTS_SERVICE_ACCOUNT"] = click.prompt(
        cyan("[METAFLOW_ARGO_EVENTS_SERVICE_ACCOUNT]")
        + " Service Account for Argo Events. ",
        default=existing_env.get("METAFLOW_ARGO_EVENTS_SERVICE_ACCOUNT", ""),
        show_default=True,
    )

    # Argo events event bus
    env["METAFLOW_ARGO_EVENTS_EVENT_BUS"] = click.prompt(
        cyan("[METAFLOW_ARGO_EVENTS_EVENT_BUS]")
        + yellow(" (optional)")
        + " Event Bus for Argo Events.",
        default=existing_env.get("METAFLOW_ARGO_EVENTS_EVENT_BUS", "default"),
        show_default=True,
    )

    # Argo events event source
    env["METAFLOW_ARGO_EVENTS_EVENT_SOURCE"] = click.prompt(
        cyan("[METAFLOW_ARGO_EVENTS_EVENT_SOURCE]") + " Event Source for Argo Events.",
        default=existing_env.get("METAFLOW_ARGO_EVENTS_EVENT_SOURCE", ""),
        show_default=True,
    )

    # Argo events event name
    env["METAFLOW_ARGO_EVENTS_EVENT"] = click.prompt(
        cyan("[METAFLOW_ARGO_EVENTS_EVENT]") + " Event name for Argo Events.",
        default=existing_env.get("METAFLOW_ARGO_EVENTS_EVENT", ""),
        show_default=True,
    )

    # Argo events webhook url
    env["METAFLOW_ARGO_EVENTS_WEBHOOK_URL"] = click.prompt(
        cyan("[METAFLOW_ARGO_EVENTS_WEBHOOK_URL]")
        + " Publicly accessible URL for Argo Events Webhook.",
        default=existing_env.get("METAFLOW_ARGO_EVENTS_WEBHOOK_URL", ""),
        show_default=True,
    )
    # Set internal URL for Argo events webhook
    env["METAFLOW_ARGO_EVENTS_INTERNAL_WEBHOOK_URL"] = click.prompt(
        cyan("[METAFLOW_ARGO_EVENTS_INTERNAL_WEBHOOK_URL]")
        + yellow(" (optional)")
        + " URL for Argo Events Webhook "
        + "(Accessible only within a Kubernetes cluster).",
        default=existing_env.get(
            "METAFLOW_ARGO_EVENTS_INTERNAL_WEBHOOK_URL",
            env["METAFLOW_ARGO_EVENTS_WEBHOOK_URL"],
        ),
        show_default=True,
    )

    return env


def configure_kubernetes(existing_env):
    empty_profile = False
    if not existing_env:
        empty_profile = True
    env = {}

    # Set K8S Namespace
    env["METAFLOW_KUBERNETES_NAMESPACE"] = click.prompt(
        cyan("[METAFLOW_KUBERNETES_NAMESPACE]")
        + yellow(" (optional)")
        + " Kubernetes Namespace ",
        default="default",
        show_default=True,
    )

    # Set K8S SA
    env["METAFLOW_KUBERNETES_SERVICE_ACCOUNT"] = click.prompt(
        cyan("[METAFLOW_KUBERNETES_SERVICE_ACCOUNT]")
        + yellow(" (optional)")
        + " Kubernetes Service Account ",
        default="default",
        show_default=True,
    )

    # Set default Docker repository for K8S jobs.
    env["METAFLOW_KUBERNETES_CONTAINER_REGISTRY"] = click.prompt(
        cyan("[METAFLOW_KUBERNETES_CONTAINER_REGISTRY]")
        + yellow(" (optional)")
        + " Default Docker image repository for K8S "
        + "jobs. If nothing is specified, "
        + "dockerhub (hub.docker.com/) is "
        + "used as default.",
        default=existing_env.get("METAFLOW_KUBERNETES_CONTAINER_REGISTRY", ""),
        show_default=True,
    )
    # Set default Docker image for K8S jobs.
    env["METAFLOW_KUBERNETES_CONTAINER_IMAGE"] = click.prompt(
        cyan("[METAFLOW_KUBERNETES_CONTAINER_IMAGE]")
        + yellow(" (optional)")
        + " Default Docker image for K8S jobs. "
        + "If nothing is specified, an appropriate "
        + "python image is used as default.",
        default=existing_env.get("METAFLOW_KUBERNETES_CONTAINER_IMAGE", ""),
        show_default=True,
    )
    # Set default Kubernetes secrets to source into pod envs
    env["METAFLOW_KUBERNETES_SECRETS"] = click.prompt(
        cyan("[METAFLOW_KUBERNETES_SECRETS]")
        + yellow(" (optional)")
        + " Comma-delimited list of secret names. Jobs will"
        " gain environment variables from these secrets. ",
        default=existing_env.get("METAFLOW_KUBERNETES_SECRETS", ""),
        show_default=True,
    )

    return env


def verify_aws_credentials(ctx):
    # Verify that the user has configured AWS credentials on their computer.
    if not click.confirm(
        "\nMetaflow relies on "
        + yellow("AWS access credentials")
        + " present on your computer to access resources on AWS."
        "\nBefore proceeding further, please confirm that you "
        "have already configured these access credentials on "
        "this computer.",
        default=True,
    ):
        echo(
            "There are many ways to setup your AWS access credentials. You "
            "can get started by following this guide: ",
            nl=False,
            fg="yellow",
        )
        echo(
            "https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-configure.html",
            fg="cyan",
        )
        ctx.abort()


def verify_azure_credentials(ctx):
    # Verify that the user has configured AWS credentials on their computer.
    if not click.confirm(
        "\nMetaflow relies on "
        + yellow("Azure access credentials")
        + " present on your computer to access resources on Azure."
        "\nBefore proceeding further, please confirm that you "
        "have already configured these access credentials on "
        "this computer.",
        default=True,
    ):
        echo(
            "There are many ways to setup your Azure access credentials. You "
            "can get started by getting familiar with the following: ",
            nl=False,
            fg="yellow",
        )
        echo("")
        echo(
            "- https://docs.microsoft.com/en-us/cli/azure/authenticate-azure-cli",
            fg="cyan",
        )
        echo(
            "- https://docs.microsoft.com/en-us/cli/azure/azure-cli-configuration",
            fg="cyan",
        )
        ctx.abort()


def verify_gcp_credentials(ctx):
    # Verify that the user has configured AWS credentials on their computer.
    if not click.confirm(
        "\nMetaflow relies on "
        + yellow("GCP access credentials")
        + " present on your computer to access resources on GCP."
        "\nBefore proceeding further, please confirm that you "
        "have already configured these access credentials on "
        "this computer.",
        default=True,
    ):
        echo(
            "There are many ways to setup your GCP access credentials. You "
            "can get started by getting familiar with the following: ",
            nl=False,
            fg="yellow",
        )
        echo("")
        echo(
            "- https://cloud.google.com/docs/authentication/provide-credentials-adc",
            fg="cyan",
        )
        ctx.abort()


@configure.command(help="Configure metaflow to access Microsoft Azure.")
@click.option(
    "--profile",
    "-p",
    default="",
    help="Configure a named profile. Activate the profile by setting "
    "`METAFLOW_PROFILE` environment variable.",
)
@click.pass_context
def azure(ctx, profile):
    # Greet the user!
    echo(
        "Welcome to Metaflow! Follow the prompts to configure your installation.\n",
        bold=True,
    )

    # Check for existing configuration.
    if not confirm_overwrite_config(profile):
        ctx.abort()

    verify_azure_credentials(ctx)

    existing_env = get_env(profile)

    env = {}
    env.update(configure_azure_datastore_and_metadata(existing_env))

    persist_env({k: v for k, v in env.items() if v}, profile)

    # Prompt user to also configure Kubernetes for compute if using azure
    if env.get("METAFLOW_DEFAULT_DATASTORE") == "azure":
        click.echo(
            "\nFinal note! Metaflow can scale your flows by "
            + yellow("executing your steps on Kubernetes.")
            + "\nYou may use Azure Kubernetes Service (AKS)"
            " or a self-managed Kubernetes cluster on Azure VMs."
            + " If/when your Kubernetes cluster is ready for use,"
            " please run 'metaflow configure kubernetes'.",
        )


@configure.command(help="Configure metaflow to access Google Cloud Platform.")
@click.option(
    "--profile",
    "-p",
    default="",
    help="Configure a named profile. Activate the profile by setting "
    "`METAFLOW_PROFILE` environment variable.",
)
@click.pass_context
def gcp(ctx, profile):
    # Greet the user!
    echo(
        "Welcome to Metaflow! Follow the prompts to configure your installation.\n",
        bold=True,
    )

    # Check for existing configuration.
    if not confirm_overwrite_config(profile):
        ctx.abort()

    verify_gcp_credentials(ctx)

    existing_env = get_env(profile)

    env = {}
    env.update(configure_gs_datastore_and_metadata(existing_env))

    persist_env({k: v for k, v in env.items() if v}, profile)

    # Prompt user to also configure Kubernetes for compute if using Google Cloud Storage
    if env.get("METAFLOW_DEFAULT_DATASTORE") == "gs":
        click.echo(
            "\nFinal note! Metaflow can scale your flows by "
            + yellow("executing your steps on Kubernetes.")
            + "\nYou may use Google Kubernetes Engine (GKE)"
            " or a self-managed Kubernetes cluster on Google Compute Engine VMs."
            + " If/when your Kubernetes cluster is ready for use,"
            " please run 'metaflow configure kubernetes'.",
        )


@configure.command(help="Configure metaflow to access self-managed AWS resources.")
@click.option(
    "--profile",
    "-p",
    default="",
    help="Configure a named profile. Activate the profile by setting "
    "`METAFLOW_PROFILE` environment variable.",
)
@click.pass_context
def aws(ctx, profile):
    # Greet the user!
    echo(
        "Welcome to Metaflow! Follow the prompts to configure your " "installation.\n",
        bold=True,
    )

    # Check for existing configuration.
    if not confirm_overwrite_config(profile):
        ctx.abort()

    verify_aws_credentials(ctx)

    existing_env = get_env(profile)
    empty_profile = False
    if not existing_env:
        empty_profile = True

    env = {}
    env.update(configure_aws_datastore_and_metadata(existing_env))

    # Configure AWS Batch for compute if using S3
    if env.get("METAFLOW_DEFAULT_DATASTORE") == "s3":
        if click.confirm(
            "\nMetaflow can scale your flows by "
            + yellow("executing your steps on AWS Batch")
            + ".\nAWS Batch is a strict requirement if you intend "
            "to schedule your flows on AWS Step Functions.\nWould "
            "you like to configure AWS Batch as your compute "
            "backend?",
            default=empty_profile or "METAFLOW_BATCH_JOB_QUEUE" in existing_env,
            abort=False,
        ):
            env.update(configure_aws_batch(existing_env))

    persist_env({k: v for k, v in env.items() if v}, profile)


@configure.command(help="Configure metaflow to use Kubernetes.")
@click.option(
    "--profile",
    "-p",
    default="",
    help="Configure a named profile. Activate the profile by setting "
    "`METAFLOW_PROFILE` environment variable.",
)
@click.pass_context
def kubernetes(ctx, profile):
    check_kubernetes_client(ctx)

    # Greet the user!
    echo(
        "Welcome to Metaflow! Follow the prompts to configure your " "installation.\n",
        bold=True,
    )

    check_kubernetes_config(ctx)

    # Check for existing configuration.
    if not confirm_overwrite_config(profile):
        ctx.abort()

    existing_env = get_env(profile)

    env = existing_env.copy()

    # We used to push user straight to S3 configuration inline.
    # Now that we support >1 cloud, it gets too complicated.
    # Therefore, we instruct the user to configure datastore first, by
    # a separate command.
    if existing_env.get("METAFLOW_DEFAULT_DATASTORE") == "local":
        click.echo(
            "\nCannot run Kubernetes with local datastore. Please run"
            " 'metaflow configure aws' or 'metaflow configure azure'."
        )
        click.Abort()

    # Configure remote metadata.
    if existing_env.get("METAFLOW_DEFAULT_METADATA") == "service":
        # Skip metadata service configuration if it is already configured
        pass
    else:
        if click.confirm(
            "\nMetaflow can use a "
            + yellow("remote Metadata Service to track")
            + " and persist flow execution metadata. \nWould you like to "
            "configure the Metadata Service?",
            default=True,
            abort=False,
        ):
            env.update(configure_metadata_service(existing_env))

    # Configure Kubernetes for compute.
    env.update(configure_kubernetes(existing_env))

    # Configure Argo Workflows Events
    if click.confirm("\nConfigure support for Argo Workflow Events?"):
        env.update(configure_argo_events(existing_env))

    persist_env({k: v for k, v in env.items() if v}, profile)



import os
import json
import time

from typing import ClassVar, Dict, Optional, TYPE_CHECKING

from metaflow.exception import MetaflowNotFound
from metaflow.metaflow_config import DEFAULT_FROM_DEPLOYMENT_IMPL


def generate_fake_flow_file_contents(
    flow_name: str, param_info: dict, project_name: Optional[str] = None
):
    params_code = ""
    for _, param_details in param_info.items():
        param_python_var_name = param_details.get(
            "python_var_name", param_details["name"]
        )
        param_name = param_details["name"]
        param_type = param_details["type"]
        param_help = param_details["description"]
        param_required = param_details["is_required"]

        if param_type == "JSON":
            params_code += (
                f"    {param_python_var_name} = Parameter('{param_name}', "
                f"type=JSONType, help='''{param_help}''', required={param_required})\n"
            )
        elif param_type == "FilePath":
            is_text = param_details.get("is_text", True)
            encoding = param_details.get("encoding", "utf-8")
            params_code += (
                f"    {param_python_var_name} = IncludeFile('{param_name}', "
                f"is_text={is_text}, encoding='{encoding}', help='''{param_help}''', "
                f"required={param_required})\n"
            )
        else:
            params_code += (
                f"    {param_python_var_name} = Parameter('{param_name}', "
                f"type={param_type}, help='''{param_help}''', required={param_required})\n"
            )

    project_decorator = f"@project(name='{project_name}')\n" if project_name else ""

    contents = f"""\
from metaflow import FlowSpec, Parameter, IncludeFile, JSONType, step, project
{project_decorator}class {flow_name}(FlowSpec):
{params_code}
    @step
    def start(self):
        self.next(self.end)
    @step
    def end(self):
        pass
if __name__ == '__main__':
    {flow_name}()
"""
    return contents


if TYPE_CHECKING:
    import metaflow
    import metaflow.runner.deployer_impl


class DeployerMeta(type):
    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)

        from metaflow.plugins import DEPLOYER_IMPL_PROVIDERS

        def _injected_method(method_name, deployer_class):
            def f(self, **deployer_kwargs):
                return deployer_class(
                    deployer_kwargs=deployer_kwargs,
                    flow_file=self.flow_file,
                    show_output=self.show_output,
                    profile=self.profile,
                    env=self.env,
                    cwd=self.cwd,
                    file_read_timeout=self.file_read_timeout,
                    **self.top_level_kwargs,
                )

            f.__doc__ = provider_class.__doc__ or ""
            f.__name__ = method_name
            return f

        for provider_class in DEPLOYER_IMPL_PROVIDERS:
            # TYPE is the name of the CLI groups i.e.
            # `argo-workflows` instead of `argo_workflows`
            # The injected method names replace '-' by '_' though.
            method_name = provider_class.TYPE.replace("-", "_")
            setattr(cls, method_name, _injected_method(method_name, provider_class))

        return cls


class Deployer(metaclass=DeployerMeta):
    """
    Use the `Deployer` class to configure and access one of the production
    orchestrators supported by Metaflow.

    Parameters
    ----------
    flow_file : str
        Path to the flow file to deploy, relative to current directory.
    show_output : bool, default True
        Show the 'stdout' and 'stderr' to the console by default.
    profile : Optional[str], default None
        Metaflow profile to use for the deployment. If not specified, the default
        profile is used.
    env : Optional[Dict[str, str]], default None
        Additional environment variables to set for the deployment.
    cwd : Optional[str], default None
        The directory to run the subprocess in; if not specified, the current
        directory is used.
    file_read_timeout : int, default 3600
        The timeout until which we try to read the deployer attribute file (in seconds).
    **kwargs : Any
        Additional arguments that you would pass to `python myflow.py` before
        the deployment command.
    """

    def __init__(
        self,
        flow_file: str,
        show_output: bool = True,
        profile: Optional[str] = None,
        env: Optional[Dict] = None,
        cwd: Optional[str] = None,
        file_read_timeout: int = 3600,
        **kwargs,
    ):
        # Convert flow_file to absolute path if it's relative
        if not os.path.isabs(flow_file):
            self.flow_file = os.path.abspath(flow_file)
        else:
            self.flow_file = flow_file

        self.show_output = show_output
        self.profile = profile
        self.env = env
        self.cwd = cwd
        self.file_read_timeout = file_read_timeout
        self.top_level_kwargs = kwargs


class TriggeredRun(object):
    """
    TriggeredRun class represents a run that has been triggered on a
    production orchestrator.
    """

    def __init__(
        self,
        deployer: "metaflow.runner.deployer_impl.DeployerImpl",
        content: str,
    ):
        self.deployer = deployer
        content_json = json.loads(content)
        self.metadata_for_flow = content_json.get("metadata")
        self.pathspec = content_json.get("pathspec")
        self.name = content_json.get("name")

    def wait_for_run(self, check_interval: int = 5, timeout: Optional[int] = None):
        """
        Wait for the `run` property to become available.

        The `run` property becomes available only after the `start` task of the triggered
        flow starts running.

        Parameters
        ----------
        check_interval: int, default: 5
            Frequency of checking for the `run` to become available, in seconds.
        timeout : int, optional, default None
            Maximum time to wait for the `run` to become available, in seconds. If
            None, wait indefinitely.

        Raises
        ------
        TimeoutError
            If the `run` is not available within the specified timeout.
        """
        start_time = time.time()
        while True:
            if self.run is not None:
                return self.run

            if timeout is not None and (time.time() - start_time) > timeout:
                raise TimeoutError(
                    "Timed out waiting for the run object to become available."
                )

            time.sleep(check_interval)

    @property
    def run(self) -> Optional["metaflow.Run"]:
        """
        Retrieve the `Run` object for the triggered run.

        Note that Metaflow `Run` becomes available only when the `start` task
        has started executing.

        Returns
        -------
        Run, optional
            Metaflow Run object if the `start` step has started executing, otherwise None.
        """
        from metaflow import Run

        try:
            return Run(self.pathspec, _namespace_check=False)
        except MetaflowNotFound:
            return None


class DeployedFlowMeta(type):
    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)
        if not bases:
            # Inject methods only in DeployedFlow and not any of its
            # subclasses
            from metaflow.plugins import DEPLOYER_IMPL_PROVIDERS

            allowed_providers = dict(
                {
                    provider.TYPE.replace("-", "_"): provider
                    for provider in DEPLOYER_IMPL_PROVIDERS
                }
            )

            def _get_triggered_run_injected_method():
                def f(
                    cls,
                    identifier: str,
                    run_id: str,
                    metadata: Optional[str] = None,
                    impl: str = DEFAULT_FROM_DEPLOYMENT_IMPL.replace("-", "_"),
                ) -> "TriggeredRun":
                    """
                    Retrieves a `TriggeredRun` object from an identifier, a run id and optional
                    metadata. The `impl` parameter specifies the deployer implementation
                    to use (like `argo-workflows`).

                    Parameters
                    ----------
                    identifier : str
                        Deployer specific identifier for the workflow to retrieve
                    run_id : str
                        Run ID for the which to fetch the triggered run object
                    metadata : str, optional, default None
                        Optional deployer specific metadata.
                    impl : str, optional, default given by METAFLOW_DEFAULT_FROM_DEPLOYMENT_IMPL
                        The default implementation to use if not specified

                    Returns
                    -------
                    TriggeredRun
                        A `TriggeredRun` object representing the triggered run corresponding
                        to the identifier and the run id.
                    """
                    if impl in allowed_providers:
                        return (
                            allowed_providers[impl]
                            .deployed_flow_type()
                            .get_triggered_run(identifier, run_id, metadata)
                        )
                    else:
                        raise ValueError(
                            f"No deployer '{impl}' exists; valid deployers are: "
                            f"{list(allowed_providers.keys())}"
                        )

                f.__name__ = "get_triggered_run"
                return f

            def _per_type_get_triggered_run_injected_method(method_name, impl):
                def f(
                    cls,
                    identifier: str,
                    run_id: str,
                    metadata: Optional[str] = None,
                ):
                    return (
                        allowed_providers[impl]
                        .deployed_flow_type()
                        .get_triggered_run(identifier, run_id, metadata)
                    )

                f.__name__ = method_name
                return f

            def _from_deployment_injected_method():
                def f(
                    cls,
                    identifier: str,
                    metadata: Optional[str] = None,
                    impl: str = DEFAULT_FROM_DEPLOYMENT_IMPL.replace("-", "_"),
                ) -> "DeployedFlow":
                    """
                    Retrieves a `DeployedFlow` object from an identifier and optional
                    metadata. The `impl` parameter specifies the deployer implementation
                    to use (like `argo-workflows`).

                    Parameters
                    ----------
                    identifier : str
                        Deployer specific identifier for the workflow to retrieve
                    metadata : str, optional, default None
                        Optional deployer specific metadata.
                    impl : str, optional, default given by METAFLOW_DEFAULT_FROM_DEPLOYMENT_IMPL
                        The default implementation to use if not specified

                    Returns
                    -------
                    DeployedFlow
                        A `DeployedFlow` object representing the deployed flow corresponding
                        to the identifier
                    """
                    if impl in allowed_providers:
                        return (
                            allowed_providers[impl]
                            .deployed_flow_type()
                            .from_deployment(identifier, metadata)
                        )
                    else:
                        raise ValueError(
                            f"No deployer '{impl}' exists; valid deployers are: "
                            f"{list(allowed_providers.keys())}"
                        )

                f.__name__ = "from_deployment"
                return f

            def _per_type_from_deployment_injected_method(method_name, impl):
                def f(
                    cls,
                    identifier: str,
                    metadata: Optional[str] = None,
                ):
                    return (
                        allowed_providers[impl]
                        .deployed_flow_type()
                        .from_deployment(identifier, metadata)
                    )

                f.__name__ = method_name
                return f

            def _list_deployed_flows_injected_method():
                def f(
                    cls,
                    flow_name: Optional[str] = None,
                    impl: str = DEFAULT_FROM_DEPLOYMENT_IMPL.replace("-", "_"),
                ):
                    """
                    List all deployed flows for the specified implementation.

                    Parameters
                    ----------
                    flow_name : str, optional, default None
                        If specified, only list deployed flows for this specific flow name.
                        If None, list all deployed flows.
                    impl : str, optional, default given by METAFLOW_DEFAULT_FROM_DEPLOYMENT_IMPL
                        The default implementation to use if not specified

                    Yields
                    ------
                    DeployedFlow
                        `DeployedFlow` objects representing deployed flows.
                    """
                    if impl in allowed_providers:
                        return (
                            allowed_providers[impl]
                            .deployed_flow_type()
                            .list_deployed_flows(flow_name)
                        )
                    else:
                        raise ValueError(
                            f"No deployer '{impl}' exists; valid deployers are: "
                            f"{list(allowed_providers.keys())}"
                        )

                f.__name__ = "list_deployed_flows"
                return f

            def _per_type_list_deployed_flows_injected_method(method_name, impl):
                def f(
                    cls,
                    flow_name: Optional[str] = None,
                ):
                    return (
                        allowed_providers[impl]
                        .deployed_flow_type()
                        .list_deployed_flows(flow_name)
                    )

                f.__name__ = method_name
                return f

            setattr(
                cls, "from_deployment", classmethod(_from_deployment_injected_method())
            )
            setattr(
                cls,
                "list_deployed_flows",
                classmethod(_list_deployed_flows_injected_method()),
            )
            setattr(
                cls,
                "get_triggered_run",
                classmethod(_get_triggered_run_injected_method()),
            )

            for impl in allowed_providers:
                from_deployment_method_name = f"from_{impl}"
                list_deployed_flows_method_name = f"list_{impl}"
                get_triggered_run_method_name = f"get_triggered_{impl}_run"

                setattr(
                    cls,
                    from_deployment_method_name,
                    classmethod(
                        _per_type_from_deployment_injected_method(
                            from_deployment_method_name, impl
                        )
                    ),
                )

                setattr(
                    cls,
                    list_deployed_flows_method_name,
                    classmethod(
                        _per_type_list_deployed_flows_injected_method(
                            list_deployed_flows_method_name, impl
                        )
                    ),
                )

                setattr(
                    cls,
                    get_triggered_run_method_name,
                    classmethod(
                        _per_type_get_triggered_run_injected_method(
                            get_triggered_run_method_name, impl
                        )
                    ),
                )

        return cls


class DeployedFlow(metaclass=DeployedFlowMeta):
    """
    DeployedFlow class represents a flow that has been deployed.

    This class is not meant to be instantiated directly. Instead, it is returned from
    methods of `Deployer`.
    """

    # This should match the TYPE value in DeployerImpl for proper stub generation
    TYPE: ClassVar[Optional[str]] = None

    def __init__(self, deployer: "metaflow.runner.deployer_impl.DeployerImpl"):
        self.deployer = deployer
        self.name = self.deployer.name
        self.flow_name = self.deployer.flow_name
        self.metadata = self.deployer.metadata
